"""
Reference management utilities for workflows and templates.

This module helps you track and validate variable references that are available in different workflow scopes. The core component is the :py:class:`Context` class, which maintains collections of available references for a specific scope.

For detailed information about references, see `Workflow Variables`_ in the Argo Workflows documentation.

.. _Workflow Variables: https://argo-workflows.readthedocs.io/en/latest/variables/
"""

__all__ = [
    "Context",
    "get_global_context",
    "get_step_context",
    "get_task_context",
    "get_template_context",
    "get_workflow_context",
]

from tugboat.references.context import Context
from tugboat.references.step import get_step_context, get_task_context
from tugboat.references.template import get_template_context
from tugboat.references.workflow import get_global_context, get_workflow_context



"""
Helper functions for generating human readable text.
"""

from __future__ import annotations

import contextlib
import functools
import io
import typing

if typing.TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence
    from typing import Any

    from pydantic import BaseModel


def join_with_and(
    items: Iterable[Any],
    *,
    quote: bool = True,
    sort: bool = True,
    fallback_string: str = "(none)",
) -> str:
    """
    Join items with "and" as the last joiner.
    On empty input, returns the fallback string.

    Arguments
    ---------
    items : Iterable[Any]
        The items to join.
    quote : bool
        Whether to quote the items.
    sort : bool
        Whether to sort the items.
    fallback_string : str
        The string to return if there are no items.
    """
    return _join(
        items=items,
        quoting=_quote if quote else str,
        last_joiner=" and ",
        sort=sort,
        fallback_string=fallback_string,
    )


def join_with_or(
    items: Iterable[Any],
    *,
    quote: bool = True,
    sort: bool = True,
    fallback_string: str = "(none)",
) -> str:
    """
    Join items with "or" as the last joiner.
    On empty input, returns the fallback string.

    Arguments
    ---------
    items : Iterable[Any]
        The items to join.
    quote : bool
        Whether to quote the items.
    sort : bool
        Whether to sort the items.
    fallback_string : str
        The string to return if there are no items.
    """
    return _join(
        items=items,
        quoting=_quote if quote else str,
        last_joiner=" or ",
        sort=sort,
        fallback_string=fallback_string,
    )


def _join(
    *,
    items: Iterable[Any],
    quoting: Callable[[Any], str],
    last_joiner: str,
    sort: bool,
    fallback_string: str,
):
    """Internal helper for :func:`join_with_and` and :func:`join_with_or`."""
    items_iter = map(quoting, items)
    if sort:
        items_iter = sorted(items_iter)
    if result := join(
        items_iter,
        last_joiner=last_joiner,
    ):
        return result
    return fallback_string


def join(
    items: Iterable[str],
    *,
    separator: str = ", ",
    last_joiner: str = ", ",
):
    """
    Join items with a separator, with a different last joiner.

    Arguments
    ---------
    items : Iterable[str]
        The items to join.

    Keyword Arguments
    -----------------
    separator : str
        The separator to use between items, by default ``, ``.
    last_joiner : str
        The string to use before the last item.

    Returns
    -------
    str
        The joined string.
    """
    with io.StringIO() as buf:
        # join items with the specified separator
        last_item = None
        for item in iter(items):
            if buf.tell():
                buf.write(separator)
            if last_item is not None:
                buf.write(last_item)
            last_item = item

        # perform the last join
        if buf.tell():
            buf.write(last_joiner)
        if last_item is not None:
            buf.write(last_item)

        return buf.getvalue()


def _quote(item: Any) -> str:
    text = str(item)
    if "'" in text:
        return f'"{text}"'
    return f"'{text}'"


def get_context_name(loc: Sequence[str | int]) -> str:
    """Get the parent context name for a location."""
    if not isinstance(loc, tuple):
        loc = tuple(loc)
    return _get_context_name(loc)


@functools.lru_cache(16)
def _get_context_name(loc: tuple[str | int, ...]) -> str:
    with contextlib.suppress(StopIteration):
        # find the first string in the parents
        parent = next(filter(lambda x: isinstance(x, str), reversed(loc)))
        return f"the '{parent}' section"
    return "current context"


def get_alias(model: BaseModel, name: str) -> str:
    """Get the alias of a field in a model."""
    field = type(model).model_fields[name]
    return field.alias or name



from __future__ import annotations

import typing

from tugboat.constraints import (
    accept_none,
    mutually_exclusive,
    require_all,
    require_non_empty,
)
from tugboat.core import hookimpl
from tugboat.references import get_workflow_context
from tugboat.types import Field
from tugboat.utils import (
    check_model_fields_references,
    check_value_references,
    find_duplicate_names,
    prepend_loc,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Literal

    from tugboat.references import Context
    from tugboat.schemas import (
        Artifact,
        Parameter,
        Template,
        Workflow,
        WorkflowTemplate,
    )
    from tugboat.types import Diagnosis


@hookimpl(specname="analyze_template")
def check_input_parameters(
    template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:
    if not template.inputs:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(template.inputs.parameters or ()):
        yield {
            "code": "TPL102",
            "loc": ("inputs", "parameters", idx, "name"),
            "summary": "Duplicate parameter name",
            "msg": f"Parameter name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each parameter
    ctx = get_workflow_context(workflow)

    for idx, param in enumerate(template.inputs.parameters or ()):
        yield from prepend_loc(
            ("inputs", "parameters", idx),
            _check_input_parameter(workflow.kind, param, ctx),
        )


def _check_input_parameter(
    kind: Literal["Workflow", "WorkflowTemplate"],
    param: Parameter,
    context: Context,
) -> Iterable[Diagnosis]:
    # generic checks for all parameters
    yield from require_non_empty(
        model=param,
        loc=(),
        fields=["name"],
    )
    yield from accept_none(
        model=param,
        loc=(),
        fields=["globalName"],
    )

    for diag in check_model_fields_references(param, context.parameters):
        match diag["code"]:
            case "VAR002":
                diag["code"] = "TPL201"
                if metadata := diag.get("ctx", {}).get("reference"):
                    ref = metadata["found:str"]
                    diag["msg"] = (
                        f"The parameter reference '{ref}' used in parameter '{param.name}' is invalid."
                    )
        yield diag

    # kind-specific checks
    if kind == "Workflow":
        yield from mutually_exclusive(
            model=param,
            loc=(),
            fields=["value", "valueFrom"],
        )

    elif kind == "WorkflowTemplate":
        yield from mutually_exclusive(
            model=param,
            loc=(),
            fields=["value", "valueFrom", "default"],
        )

        if param.value:
            yield {
                "type": "failure",
                "code": "M102",
                "loc": ("value",),
                "summary": "Found redundant field 'value'",
                "msg": (
                    """
                    The 'value' field in a WorkflowTemplate input parameter may cause issues.
                    Use 'default' instead to set a default value for the parameter.
                    """
                ),
                "input": Field("value"),
                "fix": "default",
            }

    # field-specific checks
    if param.valueFrom:
        yield from require_all(
            model=param.valueFrom,
            loc=("valueFrom",),
            fields=["configMapKeyRef"],
        )
        yield from accept_none(
            model=param.valueFrom,
            loc=("valueFrom",),
            fields=[
                "event",
                "expression",
                "jqFilter",
                "jsonPath",
                "parameter",
                "path",
                "supplied",
            ],
        )


@hookimpl(specname="analyze_template")
def check_input_artifacts(
    template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:
    if not template.inputs:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(template.inputs.artifacts or ()):
        yield {
            "code": "TPL103",
            "loc": ("inputs", "artifacts", idx, "name"),
            "summary": "Duplicate artifact name",
            "msg": f"Artifact name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each artifact;
    ctx = get_workflow_context(workflow)

    for idx, artifact in enumerate(template.inputs.artifacts or ()):
        yield from prepend_loc(
            ("inputs", "artifacts", idx),
            _check_input_artifact(template, artifact, ctx),
        )


def _check_input_artifact(
    template: Template, artifact: Artifact, context: Context
) -> Iterable[Diagnosis]:
    # generic checks for all artifacts
    yield from require_non_empty(
        model=artifact,
        loc=(),
        fields=["name"],
    )
    yield from mutually_exclusive(
        model=artifact,
        loc=(),
        fields=[
            "artifactory",
            "azure",
            "gcs",
            "git",
            "hdfs",
            "http",
            "oss",
            "raw",
            "s3",
        ],
    )
    yield from accept_none(
        model=artifact,
        loc=(),
        fields=[
            "archive",
            "archiveLogs",
            "artifactGC",
            "deleted",
            "from_",
            "fromExpression",
            "globalName",
        ],
    )

    # template type-specific: `path` should be present for container/script templates
    if template.type in ("container", "containerSet", "script"):
        if artifact.path is None:
            yield {
                "code": "M101",
                "loc": ("path",),
                "summary": "Missing required field 'path'",
                "msg": (
                    f"""
                    Field 'path' is required for artifact '{artifact.name}'.
                    Template '{template.name}' is a {template.type} template and it requires 'path' to specify file location.
                    """
                ),
            }
    else:
        if artifact.path is not None:
            yield {
                "code": "M102",
                "loc": ("path",),
                "summary": "Found redundant field 'path'",
                "msg": (
                    f"""
                    Field 'path' is invalid for artifact '{artifact.name}'.
                    Template '{template.name}' is a {template.type} template and does not accept 'path' field.
                    """
                ),
                "input": Field("path"),
            }

    # field-specific checks
    if artifact.raw:
        for diag in prepend_loc(
            ("raw", "data"),
            check_value_references(artifact.raw.data, context.parameters),
        ):
            match diag["code"]:
                case "VAR002":
                    diag["code"] = "TPL202"
                    if metadata := diag.get("ctx", {}).get("reference"):
                        ref = metadata["found:str"]
                        diag["msg"] = (
                            f"""
                            The parameter reference '{ref}' used in artifact '{artifact.name}' is invalid.
                            Note: Only parameter references are allowed here, even though this is an artifact object.
                            """
                        )
            yield diag



import os
import re


# Borrowed from CPython's `glob` module
# https://github.com/python/cpython/blob/3.13/Lib/glob.py#L267
def translate(pat, *, recursive=False, include_hidden=False, seps=None):
    """Translate a pathname with shell wildcards to a regular expression.

    If `recursive` is true, the pattern segment '**' will match any number of
    path segments.

    If `include_hidden` is true, wildcards can match path segments beginning
    with a dot ('.').

    If a sequence of separator characters is given to `seps`, they will be
    used to split the pattern into segments and match path separators. If not
    given, os.path.sep and os.path.altsep (where available) are used.
    """
    if not seps:
        if os.path.altsep:
            seps = (os.path.sep, os.path.altsep)
        else:
            seps = os.path.sep
    escaped_seps = "".join(map(re.escape, seps))
    any_sep = f"[{escaped_seps}]" if len(seps) > 1 else escaped_seps
    not_sep = f"[^{escaped_seps}]"
    if include_hidden:
        one_last_segment = f"{not_sep}+"
        one_segment = f"{one_last_segment}{any_sep}"
        any_segments = f"(?:.+{any_sep})?"
        any_last_segments = ".*"
    else:
        one_last_segment = f"[^{escaped_seps}.]{not_sep}*"
        one_segment = f"{one_last_segment}{any_sep}"
        any_segments = f"(?:{one_segment})*"
        any_last_segments = f"{any_segments}(?:{one_last_segment})?"

    results = []
    parts = re.split(any_sep, pat)
    last_part_idx = len(parts) - 1
    for idx, part in enumerate(parts):
        if part == "*":
            results.append(one_segment if idx < last_part_idx else one_last_segment)
        elif recursive and part == "**":
            if idx < last_part_idx:
                if parts[idx + 1] != "**":
                    results.append(any_segments)
            else:
                results.append(any_last_segments)
        else:
            if part:
                if not include_hidden and part[0] in "*?":
                    results.append(r"(?!\.)")
                results.extend(_translate(part, f"{not_sep}*", not_sep))
            if idx < last_part_idx:
                results.append(any_sep)
    res = "".join(results)
    return rf"(?s:{res})\Z"


# Borrowed from CPython's `fnmatch` module
# https://github.com/python/cpython/blob/3.13/Lib/fnmatch.py#L85
def _translate(pat, STAR, QUESTION_MARK):
    res = []
    add = res.append
    i, n = 0, len(pat)
    while i < n:
        c = pat[i]
        i = i + 1
        if c == "*":
            # compress consecutive `*` into one
            if (not res) or res[-1] is not STAR:
                add(STAR)
        elif c == "?":
            add(QUESTION_MARK)
        elif c == "[":
            j = i
            if j < n and pat[j] == "!":
                j = j + 1
            if j < n and pat[j] == "]":
                j = j + 1
            while j < n and pat[j] != "]":
                j = j + 1
            if j >= n:
                add("\\[")
            else:
                stuff = pat[i:j]
                if "-" not in stuff:
                    stuff = stuff.replace("\\", r"\\")
                else:
                    chunks = []
                    k = i + 2 if pat[i] == "!" else i + 1
                    while True:
                        k = pat.find("-", k, j)
                        if k < 0:
                            break
                        chunks.append(pat[i:k])
                        i = k + 1
                        k = k + 3
                    chunk = pat[i:j]
                    if chunk:
                        chunks.append(chunk)
                    else:
                        chunks[-1] += "-"
                    # Remove empty ranges -- invalid in RE.
                    for k in range(len(chunks) - 1, 0, -1):
                        if chunks[k - 1][-1] > chunks[k][0]:
                            chunks[k - 1] = chunks[k - 1][:-1] + chunks[k][1:]
                            del chunks[k]
                    # Escape backslashes and hyphens for set difference (--).
                    # Hyphens that create ranges shouldn't be escaped.
                    stuff = "-".join(
                        s.replace("\\", r"\\").replace("-", r"\-") for s in chunks
                    )
                # Escape set operations (&&, ~~ and ||).
                stuff = re.sub(r"([&~|])", r"\\\1", stuff)
                i = j + 1
                if not stuff:
                    # Empty range: never match.
                    add("(?!)")
                elif stuff == "!":
                    # Negated empty range: match any character.
                    add(".")
                else:
                    if stuff[0] == "!":
                        stuff = "^" + stuff[1:]
                    elif stuff[0] in ("^", "["):
                        stuff = "\\" + stuff
                    add(f"[{stuff}]")
        else:
            add(re.escape(c))
    assert i == n
    return res



from __future__ import annotations

import functools
import os
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field

from tugboat.schemas.arguments import Arguments
from tugboat.schemas.basic import Array, Dict
from tugboat.schemas.manifest import Manifest
from tugboat.schemas.metrics import Metrics
from tugboat.schemas.template import Template

if os.getenv("DOCUTILSCONFIG"):
    __all__ = [
        "WorkflowSpec",
        "WorkflowTemplateRef",
    ]


class WorkflowSpec(BaseModel):
    """
    `WorkflowSpec`_ is the specification of a Workflow.

    .. _WorkflowSpec: https://argo-workflows.readthedocs.io/en/latest/fields/#workflowspec
    """

    model_config = ConfigDict(extra="forbid", frozen=True)

    activeDeadlineSeconds: int | None = None
    archiveLogs: bool | None = None
    arguments: Arguments | None = None
    automountServiceAccountToken: bool | None = None
    dnsPolicy: str | None = None
    entrypoint: str | None = None
    hostNetwork: bool | None = None
    metrics: Metrics | None = None
    nodeSelector: Dict[str, str] | None = None
    onExit: str | None = None
    parallelism: int | None = None
    podPriorityClassName: str | None = None
    podSpecPatch: str | None = None
    priority: int | None = None
    schedulerName: str | None = None
    serviceAccountName: str | None = None
    shutdown: str | None = None
    suspend: bool | None = None
    templateDefaults: Template | None = None
    templates: Array[Template] | None = None
    workflowTemplateRef: WorkflowTemplateRef | None = None

    affinity: Any | None = None
    artifactGc: Any | None = Field(None, alias="artifactGC")
    artifactRepositoryRef: Any | None = None
    dnsConfig: Any | None = None
    executor: Any | None = None
    hooks: Any | None = None
    hostAliases: Array[Any] | None = None
    imagePullSecrets: Array[Any] | None = None
    podDisruptionBudget: Any | None = None
    podGc: Any | None = Field(None, alias="podGC")
    podMetadata: Any | None = None
    retryStrategy: Any | None = None
    securityContext: Any | None = None
    synchronization: Any | None = None
    tolerations: Array[Any] | None = None
    ttlStrategy: Any | None = None
    volumeClaimGc: Any | None = Field(None, alias="volumeClaimGC")
    volumeClaimTemplates: Array[Any] | None = None
    volumes: Array[Any] | None = None
    workflowMetadata: Any | None = None

    def __hash__(self):
        return hash((self.arguments, self.templates))


class WorkflowTemplateRef(BaseModel):

    model_config = ConfigDict(extra="forbid", frozen=True)

    name: str
    clusterScope: bool | None = None


class Workflow(Manifest[WorkflowSpec]):
    """
    `Workflows` are the top-level resource in Argo Workflows that define a single
    unit of work.

    .. _Workflows: https://argo-workflows.readthedocs.io/en/latest/workflow-concepts/
    """

    apiVersion: Literal["argoproj.io/v1alpha1"]  # type: ignore[reportIncompatibleVariableOverride]
    kind: Literal["Workflow"]  # type: ignore[reportIncompatibleVariableOverride]
    spec: WorkflowSpec

    @functools.cached_property
    def template_dict(self) -> dict[str, Template]:
        """
        Template name to template data model mapping.

        .. note::

           Duplicate names will be overwritten and the empty name will be ignored.
        """
        return {
            template.name: template
            for template in self.spec.templates or ()
            if template.name
        }


class WorkflowTemplate(Workflow):
    """
    `WorkflowTemplates`_ are reusable Workflow definitions stored in the cluster.

    .. _WorkflowTemplates: https://argo-workflows.readthedocs.io/en/latest/workflow-templates/
    """

    kind: Literal["WorkflowTemplate"]  # type: ignore[reportIncompatibleVariableOverride]



"""
Operators that could be used to check or report errors in the data model.
"""

from __future__ import annotations

import collections
import itertools
import typing
from collections.abc import Iterable, Sequence

from pydantic import BaseModel

from tugboat.parsers import parse_template, report_syntax_errors
from tugboat.types import Diagnosis

if typing.TYPE_CHECKING:
    from collections.abc import Container, Iterator
    from typing import Protocol, Self

    from tugboat.references.context import ReferenceCollection

    class _NamedModel(Protocol):
        name: str

    class _NullableNamedModel(Protocol):
        name: str | None

    # pyright 1.1 says non-nullable and nullable protocols are incompatible
    type NamedModel = _NamedModel | _NullableNamedModel


class prepend_loc(Iterable[Diagnosis]):
    """Prepend path to the location of each diagnosis in an iterable."""

    def __init__(self, prefix: Sequence[str | int], items: Iterable[Diagnosis] = ()):
        self.prefix = prefix
        self.items = items

    def __call__(self, diagnosis: Diagnosis) -> Diagnosis:
        diagnosis["loc"] = (*self.prefix, *diagnosis.get("loc", []))
        return diagnosis

    def __iter__(self) -> Iterator[Diagnosis]:
        return map(self, self.items)

    @classmethod
    def from_iterables(
        cls, prefix: Sequence[str | int], iterables: Iterable[Iterable[Diagnosis]]
    ) -> Self:
        return cls(prefix, itertools.chain.from_iterable(iterables))


def find_duplicate_names(items: Sequence[NamedModel]) -> Iterator[tuple[int, str]]:
    """Find and yield the indices and names of duplicate items in a sequence."""
    # count the number of times each name appears
    names = collections.defaultdict(list)
    for idx, item in enumerate(items):
        if item.name:  # skip if name is empty
            names[item.name].append(idx)

    # report any names that appear more than once
    for name, indices in names.items():
        if len(indices) > 1:
            for idx in indices:
                yield idx, name


def check_value_references(
    value: str, references: ReferenceCollection
) -> Iterator[Diagnosis]:
    """
    Check the given value for errors that are specific to Argo workflows
    variables.

    Parameters
    ----------
    value : str
        The value to check.
    references : ReferenceCollection
        The current active references.

    Yields
    ------
    Diagnosis
        A diagnosis for each error found.
    """
    doc = parse_template(value)
    yield from report_syntax_errors(doc)

    for node, ref in doc.iter_references():
        if ref in references:
            continue

        ref_str = ".".join(ref)
        metadata = {
            "found": ref,
            "found:str": ref_str,
        }

        closest = references.find_closest(ref)
        if closest:
            metadata["closest"] = closest
            metadata["closest:str"] = ".".join(closest)

        yield {
            "code": "VAR002",
            "loc": (),
            "summary": "Invalid reference",
            "msg": f"The used reference '{ref_str}' is invalid.",
            "input": str(node),
            "fix": node.format(closest),
            "ctx": {"reference": metadata},
        }


def check_model_fields_references(
    model: BaseModel, references: ReferenceCollection, *, exclude: Container[str] = ()
) -> Iterator[Diagnosis]:
    """
    Check the fields of the given model for errors that are specific to Argo
    workflows variables.

    Parameters
    ----------
    model : BaseModel
        The model to check. This function will check all fields of the model
        that are of type str.
    references : ReferenceCollection
        The current active references.
    exclude : Container[str]
        The fields to exclude from the check.

    Yields
    ------
    Diagnosis
        A diagnosis for each error found.
    """

    def _check(item):
        if isinstance(item, str):
            yield from check_value_references(item, references)

        elif isinstance(item, BaseModel):
            for field, value in item:
                yield from prepend_loc((field,), _check(value))

        elif isinstance(item, Sequence):
            for idx, child in enumerate(item):
                yield from prepend_loc((idx,), _check(child))

    for field, value in model:
        if field not in exclude:
            yield from prepend_loc((field,), _check(value))



from __future__ import annotations

import itertools
import typing

import tugboat.analyzers.workflow
from tugboat.analyzers.kubernetes import check_resource_name
from tugboat.constraints import require_exactly_one
from tugboat.core import hookimpl
from tugboat.types import Field
from tugboat.utils import prepend_loc

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.schemas import WorkflowTemplate
    from tugboat.types import Diagnosis


@hookimpl(specname="analyze_workflow_template")
def check_metadata(workflow_template: WorkflowTemplate) -> Iterator[Diagnosis]:
    yield from require_exactly_one(
        model=workflow_template.metadata,
        loc=("metadata",),
        fields=["name", "generateName"],
    )

    if workflow_template.metadata.name:
        yield from prepend_loc(
            ["metadata", "name"],
            check_resource_name(workflow_template.metadata.name, max_length=63),
        )

    if workflow_template.metadata.generateName:
        yield {
            "code": "WT001",
            "loc": ("metadata", "generateName"),
            "summary": "Use strict name",
            "msg": "Use a strict name instead of a generateName.",
            "input": Field("generateName"),
            "fix": "name",
        }


@hookimpl(specname="analyze_workflow_template")
def check_spec(workflow_template: WorkflowTemplate) -> Iterator[Diagnosis]:
    yield from require_exactly_one(
        model=workflow_template.spec,
        loc=("spec",),
        fields=["templates", "workflowTemplateRef"],
    )


@hookimpl(specname="analyze_workflow_template")
def check_entrypoint(workflow_template: WorkflowTemplate) -> Iterator[Diagnosis]:
    for diagnosis in tugboat.analyzers.workflow.check_entrypoint(workflow_template):
        match diagnosis["code"]:
            case "WF201":
                diagnosis["code"] = "WT201"
        yield diagnosis


@hookimpl(specname="analyze_workflow_template")
def check_arguments(workflow_template: WorkflowTemplate) -> Iterator[Diagnosis]:
    for diagnosis in itertools.chain(
        tugboat.analyzers.workflow.check_argument_parameters(workflow_template),
        tugboat.analyzers.workflow.check_argument_artifacts(workflow_template),
    ):
        match diagnosis["code"]:
            case "WF101":
                diagnosis["code"] = "WT101"
            case "WF102":
                diagnosis["code"] = "WT102"
        yield diagnosis


@hookimpl(specname="analyze_workflow_template")
def check_metrics(workflow_template: WorkflowTemplate) -> Iterator[Diagnosis]:
    for diagnosis in tugboat.analyzers.workflow.check_metrics(workflow_template):
        match diagnosis["code"]:
            case "WF301":
                diagnosis["code"] = "WT301"
            case "WF302":
                diagnosis["code"] = "WT302"
            case "WF303":
                diagnosis["code"] = "WT303"
        yield diagnosis



__all__ = [
    "Document",
    "Node",
    "SimpleReferenceTag",
    "Unexpected",
]

from tugboat.parsers.ast.argo_template import Document, SimpleReferenceTag
from tugboat.parsers.ast.core import Node, Unexpected



from __future__ import annotations

import logging
import typing
import warnings
from pathlib import Path

from tugboat.types import GlobPath

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from os import PathLike


logger = logging.getLogger(__name__)


def gather_paths(
    includes: Iterable[Path | GlobPath],
    excludes: Iterable[Path | GlobPath],
    follow_symlinks: bool = False,
) -> list[Path]:
    """
    Gather paths from the given include patterns and exclude patterns.
    """
    exclude_list = PathList(excludes)

    output = set()
    for item in includes:
        for file in yield_files(item, follow_symlinks=follow_symlinks):
            if file not in exclude_list:
                output.add(file)

    return list(output)


class PathList:

    def __init__(self, paths: Iterable[Path | GlobPath]):
        self.files = {
            path for path in paths if isinstance(path, Path) and path.is_file()
        }
        self.dirs = {path for path in paths if isinstance(path, Path) and path.is_dir()}

        # note - must be a list, need to run __eq__ on each item
        self.patterns = [path for path in paths if isinstance(path, GlobPath)]

    def __contains__(self, p: PathLike) -> bool:
        path = Path(p).resolve()
        if path in self.files:
            return True
        if any(path.is_relative_to(dir_) for dir_ in self.dirs):
            return True
        if path in self.patterns:
            return True
        return False


def yield_files(path: Path | GlobPath, follow_symlinks: bool = False) -> Iterator[Path]:
    if isinstance(path, GlobPath):
        logger.debug("Include glob pattern: %s", path)
        if follow_symlinks:
            warnings.warn(
                "Symlinks are not followed when using path patterns.",
                UserWarning,
                stacklevel=1,
            )
        yield from path.iglob(recursive=True, include_hidden=True)

    elif path.is_file():
        logger.debug("Include file: %s", path)
        yield path

    elif path.is_dir():
        logger.debug("Include directory: %s", path)
        for root, _, files in path.walk(follow_symlinks=follow_symlinks):
            for file in map(root.joinpath, files):
                if file.suffix in (".yaml", ".yml"):
                    yield file



from __future__ import annotations

import re
import typing

from tugboat.constraints import require_exactly_one, require_non_empty
from tugboat.utils import check_model_fields_references

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.references import Context
    from tugboat.schemas.metrics import Prometheus
    from tugboat.types import Diagnosis


def check_prometheus(prometheus: Prometheus, context: Context) -> Iterator[Diagnosis]:
    """
    Check :py:class:``tugboat.schemas.metrics.Prometheus`` for errors.
    """

    yield from require_non_empty(
        model=prometheus,
        loc=(),
        fields=["name", "help"],
    )
    yield from require_exactly_one(
        model=prometheus,
        loc=(),
        fields=["counter", "gauge", "histogram"],
    )
    yield from check_model_fields_references(
        model=prometheus,
        references=context.parameters,
    )

    if prometheus.name:
        if not re.fullmatch(r"[a-zA-Z_][a-zA-Z0-9_]*", prometheus.name):
            yield {
                "type": "failure",
                "code": "internal:invalid-metric-name",
                "loc": ("name",),
                "summary": "Invalid metric name",
                "msg": f"""
                    Metric name '{prometheus.name}' is invalid.
                    Argo Workflows metric names must start with an alphabetic character and can only include alphanumeric characters and underscores (_).
                    """,
                "input": prometheus.name,
            }

        if len(prometheus.name) > 255:
            yield {
                "type": "failure",
                "code": "internal:invalid-metric-name",
                "loc": ("name",),
                "summary": "Invalid metric name",
                "msg": """
                    Metric name is too long.
                    Metric names must be less than 256 characters.
                    """,
                "input": prometheus.name,
            }

    for idx, label in enumerate(prometheus.labels or ()):
        if not re.fullmatch(r"[a-zA-Z_][a-zA-Z0-9_]*", label.key):
            yield {
                "type": "failure",
                "code": "internal:invalid-metric-label-name",
                "loc": ("labels", idx, "key"),
                "summary": "Invalid metric label",
                "msg": f"""
                    Label name '{label.key}' in metric '{prometheus.name}' is invalid.
                    Prometheus label names must start with an alphabetic character and can only contain alphanumeric characters and underscores (_).
                    """,
                "input": label.key,
            }

        if label.key.startswith("__"):
            yield {
                "type": "failure",
                "code": "internal:invalid-metric-label-name",
                "loc": ("labels", idx, "key"),
                "summary": "Invalid metric label",
                "msg": f"""
                    Label name '{label.key}' in metric '{prometheus.name}' is invalid.
                    Label names starts with '__' are reserved for Prometheus internal use.
                    """,
                "input": label.key,
            }

        if label.value == "":
            yield {
                "type": "warning",
                "code": "internal:invalid-metric-label-value",
                "loc": ("labels", idx, "value"),
                "summary": "Redundant metric label",
                "msg": f"""
                    Label value for label '{label.key}' in metric '{prometheus.name}' is empty.
                    Labels with an empty label value are considered equivalent to labels that do not exist.
                    """,
            }



"""
This module provides some generic constraints that can be used on linting models.

All functions in this module are generators that yield :py:class:`tugboat.Diagnosis`
objects when a constraint is not met. These functions can be used in analysis
hooks, yielding results using the :py:keyword:`yield from <yield>` syntax.

A typical usage of these functions is as follows:

.. code-block:: python

   from tugboat import hookimpl
   from tugboat.constraints import require_exactly_one

   @hookimpl
   def analyze_workflow(workflow: Workflow) -> Iterator[Diagnosis]:
       yield from require_exactly_one(
           model=workflow.metadata,
           loc=("metadata",),
           fields=["name", "generateName"],
       )
"""

from __future__ import annotations

import functools
import typing

from tugboat.types import Field
from tugboat.utils import get_alias, get_context_name, join_with_and, join_with_or

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from typing import Any

    from pydantic import BaseModel

    from tugboat.types import Diagnosis


def accept_none(
    *, model: BaseModel, loc: Sequence[str | int], fields: Iterable[str]
) -> Iterator[Diagnosis]:
    """
    Check if all the specified fields are not set.

    Yield
    -----
    :rule:`m102` for each unexpected field.
    """
    for field_name in fields:
        if getattr(model, field_name, None) is not None:
            field_alias = get_alias(model, field_name)
            yield {
                "type": "failure",
                "code": "M102",
                "loc": (*loc, field_alias),
                "summary": f"Found redundant field '{field_alias}'",
                "msg": f"Field '{field_alias}' is not valid within {get_context_name(loc)}.",
                "input": Field(field_alias),
            }


def mutually_exclusive(
    *, model: Any, loc: Sequence[str | int], fields: Iterable[str]
) -> Iterator[Diagnosis]:
    """
    Ensures that at most one of the specified fields in the model is set, but
    does not require any of them to be set.

    Yield
    -----
    :rule:`m201` when more than one were set.
    """
    fields_with_values = [
        field for field in fields if getattr(model, field, None) is not None
    ]
    if len(fields_with_values) <= 1:
        return

    get_alias_ = functools.partial(get_alias, model)
    exclusive_fields = join_with_and(map(get_alias_, fields))
    fields_with_values = sorted(map(get_alias_, fields_with_values))

    for field_alias in fields_with_values:
        yield {
            "type": "failure",
            "code": "M201",
            "loc": (*loc, field_alias),
            "summary": "Mutually exclusive field set",
            "msg": f"Field {exclusive_fields} are mutually exclusive.",
            "input": Field(field_alias),
        }


def require_all(
    *, model: Any, loc: Sequence[str | int], fields: Iterable[str]
) -> Iterator[Diagnosis]:
    """
    Requires that all of the specified fields in the model are set.

    Yield
    -----
    :rule:`m101` when any of the fields are absent.
    """
    for field_name in fields:
        if getattr(model, field_name, None) is None:
            field_alias = get_alias(model, field_name)
            context_name = get_context_name(loc)
            yield {
                "type": "failure",
                "code": "M101",
                "loc": (*loc, field_alias),
                "summary": f"Missing required field '{field_alias}'",
                "msg": f"Field '{field_alias}' is required in {context_name} but missing.",
            }


def require_non_empty(
    *, model: Any, loc: Sequence[str | int], fields: Iterable[str]
) -> Iterator[Diagnosis]:
    """
    Requires that all of the specified fields are set and not empty.

    Yield
    -----
    :rule:`m101` when any of the fields are absent.
    :rule:`m202` when any of the fields are empty.
    """
    # check absent
    yield from require_all(model=model, loc=loc, fields=fields)

    # check empty
    for field_name in fields:
        field_value = getattr(model, field_name, None)
        if not field_value:
            if field_value is None or field_value is False:
                continue  # false positive

            field_alias = get_alias(model, field_name)
            context_name = get_context_name(loc)
            yield {
                "type": "failure",
                "code": "M202",
                "loc": (*loc, field_alias),
                "summary": f"Missing input in field '{field_alias}'",
                "msg": f"Field '{field_alias}' is required in {context_name} but is currently empty.",
            }


def require_exactly_one(
    *, model: Any, loc: Sequence[str | int], fields: Iterable[str]
) -> Iterator[Diagnosis]:
    """
    Requires that exactly one of the specified fields in the model is set.

    Yield
    -----
    :rule:`m101` when none of the fields are set.
    :rule:`m201` when more than one were set.
    """
    # check if any of the fields are set
    any_field_set = False
    for field_name in fields:
        value = getattr(model, field_name, None)
        if value is not None:
            any_field_set = True
            break

    if not any_field_set:
        required_fields = map(functools.partial(get_alias, model), fields)
        yield {
            "type": "failure",
            "code": "M101",
            "loc": tuple(loc),
            "summary": "Missing required field",
            "msg": f"""
                    Missing required field for {get_context_name(loc)}.
                    One of the following fields is required: {join_with_or(required_fields)}.
                    """,
        }

    yield from mutually_exclusive(model=model, loc=loc, fields=fields)



from __future__ import annotations

import os
import typing

from pydantic import BaseModel, ConfigDict, field_validator
from pydantic_core import PydanticCustomError

from tugboat.schemas.basic import Array, ConfigKeySelector, Empty

if typing.TYPE_CHECKING:
    from typing import Any

if os.getenv("DOCUTILSCONFIG"):
    __all__ = ["ValueFrom"]


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class Parameter(_BaseModel):
    """
    `Parameter`_ indicate a passed string parameter to a service template with
    an optional default value.

    .. _Parameter: https://argo-workflows.readthedocs.io/en/latest/fields/#parameter
    """

    default: str | None = None
    description: str | None = None
    enum: Array[str] | None = None
    globalName: str | None = None
    name: str
    valueFrom: ValueFrom | None = None

    value: bool | int | str | None = None
    """
    The literal value to use for the parameter.

    .. note::

       While documented as a string type in the official Argo Workflows documentation,
       this field also accepts boolean and integer values in practice.
    """

    def __hash__(self):
        return hash((repr(self.value), self.valueFrom))

    @field_validator("value", mode="plain")
    @classmethod
    def _validate_value(cls, value: Any) -> bool | int | str | None:
        if value is None:
            return value
        if isinstance(value, bool | int | str):
            return value
        raise PydanticCustomError("parameter_value_type_error", "")


class ValueFrom(_BaseModel):
    """
    `ValueFrom`_ describes a location in which to obtain the value to a parameter.

    .. _ValueFrom: https://argo-workflows.readthedocs.io/en/latest/fields/#valuefrom
    """

    configMapKeyRef: ConfigKeySelector | None = None
    default: str | None = None
    event: str | None = None
    expression: str | None = None
    jqFilter: str | None = None
    jsonPath: str | None = None
    parameter: str | None = None
    path: str | None = None
    supplied: Empty | None = None



from __future__ import annotations

import collections
import copy
import functools
import threading
import typing
from typing import overload

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from tugboat.references.context import Context


class cache:
    """
    Decorator to cache the results of a function.

    This decorator is like :py:func:`functools.lru_cache`, but it returns a copy
    of the cached value instead of the original value.
    """

    def __init__(self, max_size: int):
        self.max_size = max_size
        self.store = LruDict(max_size=max_size)

    def __call__[**P](self, func: Callable[P, Context]) -> Callable[P, Context]:
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            assert not kwargs
            if (value := self.store.get(args)) is None:
                self.store[args] = value = func(*args, **kwargs)
            return copy.deepcopy(value)

        return _wrapper


class LruDict[TK, TV](collections.OrderedDict[TK, TV]):
    """A dict that implements LRU cache"""

    def __init__(self, max_size=128):
        super().__init__()
        self.max_size = max_size
        self._lock = threading.RLock()

    def __getitem__(self, key: TK) -> TV:
        with self._lock:
            value = super().__getitem__(key)
            self.move_to_end(key)
            return value

    def __setitem__(self, key: TK, value: TV) -> None:
        with self._lock:
            super().__setitem__(key, value)
            self.move_to_end(key)
            if len(self) > self.max_size:
                first = next(iter(self))
                self.__delitem__(first)

    def __delitem__(self, key: TK) -> None:
        with self._lock:
            super().__delitem__(key)

    @overload
    def get(self, key: TK) -> TV | None: ...
    @overload
    def get(self, key: TK, default: TV) -> TV: ...
    @overload
    def get[T](self, key: TK, default: T) -> TV | T: ...

    def get[T](self, key: TK, default: T | None = None) -> TV | T | None:  # type: ignore[reportIncompatibleMethodOverride]
        try:
            return self[key]
        except KeyError:
            return default



from __future__ import annotations

import enum
import io
import re
import typing
import urllib.parse
from dataclasses import dataclass

from tugboat.console.formatters.base import OutputFormatter

if typing.TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from re import Match
    from typing import TextIO

    from tugboat.engine import DiagnosisModel


class Command(enum.StrEnum):
    """GitHub Actions Workflow commands"""

    Warning = "warning"
    Error = "error"


class GitHubFormatter(OutputFormatter):
    """
    Outputs diagnoses as GitHub Actions workflow commands.

    See also
    --------
    `Workflow commands for GitHub Actions
    <https://docs.github.com/en/actions/reference/workflows-and-actions/workflow-commands>`_
    """

    def __init__(self):
        super().__init__()
        self.commands: list[IssueCommand] = []

    def update(self, *, content: str, diagnoses: Sequence[DiagnosisModel]) -> None:
        for diagnosis in diagnoses:
            match diagnosis.type:
                case "error" | "failure":
                    command = Command.Error
                case "warning":
                    command = Command.Warning

            file = None
            if diagnosis.extras.file:
                file = diagnosis.extras.file.filepath

            title = diagnosis.code
            if diagnosis.summary:
                title += f" ({diagnosis.summary})"

            self.commands.append(
                IssueCommand(
                    command=command,
                    file=file,
                    line=diagnosis.line,
                    title=title,
                    message=diagnosis.msg,
                )
            )

    def dump(self, stream: TextIO):
        for cmd in self.commands:
            stream.write(str(cmd))
            stream.write("\n")


@dataclass
class IssueCommand:

    command: Command
    file: str | None
    line: int
    title: str
    message: str

    def __str__(self) -> str:
        # https://github.com/actions/toolkit/blob/%40actions/cache%404.0.0/packages/core/src/command.ts#L53

        def _properties() -> Iterator[tuple[str, str]]:
            if self.file:
                yield "file", escape_properties(self.file)
            yield "line", str(self.line)
            yield "title", escape_properties(self.title)

        with io.StringIO() as buf:
            # command
            buf.write("::")
            buf.write(self.command.value)

            # properties
            first = True
            for key, value in _properties():
                if first:
                    buf.write(" ")
                else:
                    buf.write(",")
                buf.write(key)
                buf.write("=")
                buf.write(value)
                first = False

            buf.write("::")

            # message
            buf.write(escape_message(self.message))

            return buf.getvalue()


def escape_properties(s: str) -> str:
    # https://github.com/actions/toolkit/blob/%40actions/cache%404.0.0/packages/core/src/command.ts#L87
    return re.sub(r"[\n\r%,:]+", _quote, s)


def escape_message(s: str) -> str:
    # https://github.com/actions/toolkit/blob/%40actions/cache%404.0.0/packages/core/src/command.ts#L80
    return re.sub(r"[\n\r%]+", _quote, s)


def _quote(m: Match[str]) -> str:
    return urllib.parse.quote(m.group(0))



from __future__ import annotations

import os
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic_core import PydanticCustomError

from tugboat.schemas.basic import (
    Array,
    ConfigKeySelector,
    Empty,
    KeyValuePair,
    NameValuePair,
    PodMetadata,
)

if os.getenv("DOCUTILSCONFIG"):
    __all__ = [
        "ArchiveStrategy",
        "ArtifactGc",
        "ArtifactoryArtifact",
        "AzureArtifact",
        "BasicAuth",
        "ClientCertAuth",
        "CreateS3BucketOptions",
        "GcsArtifact",
        "GitArtifact",
        "HdfsArtifact",
        "HttpArtifact",
        "HttpAuth",
        "OAuth2Auth",
        "OssArtifact",
        "OssLifecycleRule",
        "PluginArtifact",
        "RawArtifact",
        "S3Artifact",
        "S3EncryptionOptions",
        "TarStrategy",
    ]


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class Artifact(_BaseModel):
    """
    `Artifact`_ indicates an artifact to place at a specified path.

    .. _Artifact: https://argo-workflows.readthedocs.io/en/latest/fields/#artifact
    """

    archive: ArchiveStrategy | None = None
    archiveLogs: bool | None = None
    artifactGc: ArtifactGc | None = Field(None, alias="artifactGC")
    artifactory: ArtifactoryArtifact | None = None
    azure: AzureArtifact | None = None
    deleted: bool | None = None
    from_: str | None = Field(None, alias="from")
    fromExpression: str | None = None
    gcs: GcsArtifact | None = None
    git: GitArtifact | None = None
    globalName: str | None = None
    hdfs: HdfsArtifact | None = None
    http: HttpArtifact | None = None
    mode: int | None = None
    name: str
    optional: bool = False
    oss: OssArtifact | None = None
    path: str | None = None
    plugin: PluginArtifact | None = None
    raw: RawArtifact | None = None
    recurseMode: bool | None = None
    s3: S3Artifact | None = None
    subPath: str | None = None

    value: Literal[None] = None
    """Not a valid field for artifact. Preserve for validation purposes.
    """

    def __hash__(self) -> int:
        return hash(
            (
                self.artifactory,
                self.azure,
                self.gcs,
                self.git,
                self.hdfs,
                self.http,
                self.oss,
                self.raw,
                self.s3,
            )
        )

    @field_validator("value", mode="plain")
    @classmethod
    def _validate_(cls, value: Any) -> Literal[None]:
        if value is not None:
            raise PydanticCustomError(
                "artifact_prohibited_value_field",
                "Field 'value' is not a valid field for artifact. Use 'raw' artifact type instead.",
            )


# ----------------------------------------------------------------------------
# archive
# ----------------------------------------------------------------------------
class ArchiveStrategy(_BaseModel):
    none: Empty | None = None
    tar: TarStrategy | None = None
    zip: Empty | None = None


class TarStrategy(_BaseModel):
    compressionLevel: int | None = None


# ----------------------------------------------------------------------------
# artifactGC
# ----------------------------------------------------------------------------
class ArtifactGc(_BaseModel):
    podMetadata: PodMetadata | None = None
    serviceAccountName: str | None = None
    strategy: str


# ----------------------------------------------------------------------------
# artifactory
# ----------------------------------------------------------------------------
class ArtifactoryArtifact(_BaseModel):
    passwordSecret: ConfigKeySelector
    url: str
    usernameSecret: ConfigKeySelector


# ----------------------------------------------------------------------------
# azure
# ----------------------------------------------------------------------------
class AzureArtifact(_BaseModel):
    accountKeySecret: ConfigKeySelector
    blob: str
    container: str
    endpoint: str
    useSdkCreds: bool | None = Field(None, alias="useSDKCreds")


# ----------------------------------------------------------------------------
# gcs
# ----------------------------------------------------------------------------
class GcsArtifact(_BaseModel):
    bucket: str
    key: str
    serviceAccountKeySecret: ConfigKeySelector


# ----------------------------------------------------------------------------
# git
# ----------------------------------------------------------------------------
class GitArtifact(_BaseModel):
    branch: str | None = None
    depth: int | None = None
    disableSubmodules: bool | None = None
    fetch: Array[str] | None = None
    insecureIgnoreHostKey: bool | None = None
    insecureSkipTls: bool | None = Field(None, alias="insecureSkipTLS")
    passwordSecret: ConfigKeySelector | None = None
    repo: str
    revision: str | None = None
    singleBranch: bool | None = None
    sshPrivateKeySecret: ConfigKeySelector | None = None
    usernameSecret: ConfigKeySelector | None = None


# ----------------------------------------------------------------------------
# hdfs
# ----------------------------------------------------------------------------
class HdfsArtifact(_BaseModel):
    addresses: Array[str]
    dataTransferProtection: str | None = None
    force: bool | None = None
    hdfsUser: str
    krbCCacheSecret: ConfigKeySelector | None = None
    krbConfigConfigMap: ConfigKeySelector | None = None
    krbKeytabSecret: ConfigKeySelector | None = None
    krbRealm: str | None = None
    krbServicePrincipalName: str | None = None
    krbUsername: str | None = None
    path: str


# ----------------------------------------------------------------------------
# http
# ----------------------------------------------------------------------------
class HttpArtifact(_BaseModel):
    auth: HttpAuth | None = None
    headers: Array[NameValuePair] | None = None
    url: str


class BasicAuth(_BaseModel):
    passwordSecret: ConfigKeySelector
    usernameSecret: ConfigKeySelector


class ClientCertAuth(_BaseModel):
    clientCertSecret: ConfigKeySelector | None = None
    clientKeySecret: ConfigKeySelector


class HttpAuth(_BaseModel):
    basic: BasicAuth | None = None
    clientCert: ClientCertAuth | None = None
    oauth2: OAuth2Auth | None = None


class OAuth2Auth(_BaseModel):
    clientIdSecret: ConfigKeySelector = Field(alias="clientIDSecret")
    clientSecretSecret: ConfigKeySelector
    endpointParams: Array[KeyValuePair] | None = None
    scopes: Array[str] | None = None
    tokenUrlSecret: ConfigKeySelector | None = Field(None, alias="tokenURLSecret")


# ----------------------------------------------------------------------------
# oss
# ----------------------------------------------------------------------------
class OssArtifact(_BaseModel):
    accessKeySecret: ConfigKeySelector | None = None
    bucket: str
    createBucketIfNotPresent: bool | None = None
    endpoint: str | None = None
    key: str
    lifecycleRule: OssLifecycleRule | None = None
    secretKeySecret: ConfigKeySelector | None = None
    securityToken: str | None = None
    useSdkCreds: bool | None = Field(None, alias="useSDKCreds")


class OssLifecycleRule(_BaseModel):
    markDeletionAfterDays: int | None = None
    markInfrequentAccessAfterDays: int | None = None


# ----------------------------------------------------------------------------
# plugin
# ----------------------------------------------------------------------------
class PluginArtifact(_BaseModel):
    """
    `PluginArtifact`_ is the location of an artifact plugin.

    .. _PluginArtifact: https://argo-workflows.readthedocs.io/en/latest/fields/#pluginartifact
    """

    configuration: str | None = None
    connectionTimeoutSeconds: int | None = None
    key: str | None = None
    name: str | None


# ----------------------------------------------------------------------------
# raw
# ----------------------------------------------------------------------------
class RawArtifact(_BaseModel):
    data: str


# ----------------------------------------------------------------------------
# s3
# ----------------------------------------------------------------------------
class S3Artifact(_BaseModel):
    accessKeySecret: ConfigKeySelector | None = None
    bucket: str | None = None
    caSecret: ConfigKeySelector | None = None
    createBucketIfNotPresent: CreateS3BucketOptions | None = None
    encryptionOptions: S3EncryptionOptions | None = None
    endpoint: str | None = None
    insecure: bool | None = None
    key: str
    region: str | None = None
    roleArn: str | None = Field(None, alias="roleARN")
    secretKeySecret: ConfigKeySelector | None = None
    sessionTokenSecret: ConfigKeySelector | None = None
    useSdkCreds: bool | None = Field(None, alias="useSDKCreds")


class CreateS3BucketOptions(_BaseModel):
    objectLocking: bool


class S3EncryptionOptions(_BaseModel):
    enableEncryption: bool | None = None
    kmsEncryptionContext: str | None = None
    kmsKeyId: str | None = None
    serverSideCustomerKeySecret: ConfigKeySelector | None = None



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.basic import Array, ConfigKeySelector


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class Probe(_BaseModel):
    exec: ExecAction | None = None
    failureThreshold: int | None = None
    grpc: GrpcAction | None = None
    httpGet: HttpGetAction | None = None
    initialDelaySeconds: int | None = None
    periodSeconds: int | None = None
    successThreshold: int | None = None
    tcpSocket: TcpSocketAction | None = None
    terminationGracePeriodSeconds: int | None = None
    timeoutSeconds: int | None = None


class ExecAction(_BaseModel):
    """
    `ExecAction`_ describes a "run in container" action.

    .. _ExecAction: https://argo-workflows.readthedocs.io/en/latest/fields/#execaction
    """

    command: Array[str]


class GrpcAction(_BaseModel):
    port: int
    service: str | None = None


class HttpGetAction(_BaseModel):
    """
    `HttpGetAction`_ describes an action based on HTTP Get requests.

    .. _HttpGetAction: https://argo-workflows.readthedocs.io/en/latest/fields/#httpgetaction
    """

    host: str | None = None
    httpHeaders: Array[HttpHeader] | None = None
    path: str
    port: int | str
    scheme: Literal["HTTP", "HTTPS"] | None = None


class HttpHeader(_BaseModel):
    name: str
    value: str | None = None
    valueFrom: HttpHeaderSource | None = None


class HttpHeaderSource(_BaseModel):
    secretKeyRef: ConfigKeySelector


class TcpSocketAction(_BaseModel):
    """
    `TcpSocketAction`_ describes an action based on opening a socket.

    .. _TcpSocketAction: https://argo-workflows.readthedocs.io/en/latest/fields/#tcpsocketaction
    """

    host: str | None = None
    port: int | str



from __future__ import annotations

import collections
import datetime
import logging
import typing
from xml.etree.ElementTree import Element, ElementTree, SubElement

from tugboat.console.formatters.base import OutputFormatter

if typing.TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TextIO

    from tugboat.engine import (
        DiagnosisModel,
        FilesystemMetadata,
        HelmMetadata,
        ManifestMetadata,
    )


logger = logging.getLogger(__name__)


class JUnitFormatter(OutputFormatter):

    def __init__(self):
        super().__init__()
        self.testsuites: dict[tuple[str | None, str | None], ElementTestSuite] = {}

    def update(self, *, content: str, diagnoses: Sequence[DiagnosisModel]) -> None:
        for diagnosis in diagnoses:
            # find or create the appropriate testsuite element
            file_path = None
            if diagnosis.extras.file and not diagnosis.extras.file.is_stdin:
                file_path = diagnosis.extras.file.filepath

            manifest_name = None
            if diagnosis.extras.manifest:
                manifest_name = f"{diagnosis.extras.manifest.fqk}/{diagnosis.extras.manifest.name!r}"

            key = (file_path, manifest_name)
            if key not in self.testsuites:
                self.testsuites[key] = ElementTestSuite(
                    filesystem=diagnosis.extras.file,
                    helm=diagnosis.extras.helm,
                    manifest=diagnosis.extras.manifest,
                )

            testsuite = self.testsuites[key]

            # create a testcase element for this diagnosis
            testcase = ElementTestCase(diagnosis)
            testsuite.append(testcase)

    def dump(self, stream: TextIO) -> None:
        # create root <testsuites> element
        now = datetime.datetime.now().astimezone()
        attrib = {
            "name": "tugboat",
            "timestamp": now.isoformat(),
        }

        counter = collections.Counter()
        for testsuite in self.testsuites.values():
            for stat_name, count in testsuite.counter.items():
                counter[stat_name] += count

        for stat_name, count in counter.items():
            attrib[stat_name] = str(count)

        testsuites = Element("testsuites", attrib)

        # append all <testsuite> elements
        for testsuite in self.testsuites.values():
            testsuites.append(testsuite)

        # serialize <testsuites> element
        tree = ElementTree(testsuites)
        tree.write(stream, encoding="unicode", xml_declaration=True)


class ElementTestSuite(Element):
    """<testsuite> element = a manifest"""

    def __init__(
        self,
        *,
        filesystem: FilesystemMetadata | None = None,
        helm: HelmMetadata | None = None,
        manifest: ManifestMetadata | None = None,
    ):
        # create <testsuite> element
        now = datetime.datetime.now().astimezone()
        attrib = {
            "timestamp": now.isoformat(),
        }

        if manifest:
            try:
                attrib["name"] = manifest.fqkn
            except ValueError:
                attrib["name"] = f"{manifest.fqk}/<unnamed>"

        if filesystem and not filesystem.is_stdin:
            attrib["file"] = filesystem.filepath

        super().__init__("testsuite", attrib)

        # internal counter for statistics
        self.counter = collections.Counter()

        # create <properties> element
        properties = ElementProperties()
        self.append(properties)

        if helm:
            properties.add_property("string:helm-chart", helm.chart)
            properties.add_property("string:helm-template", helm.template)

        if manifest:
            properties.add_property("string:manifest-kind", manifest.fqk)

            if manifest.name:
                properties.add_property("string:manifest-name", manifest.name)

    def append(self, subelement):
        if isinstance(subelement, ElementTestCase):
            self.counter[subelement.stat_name] += 1
            for stat_name, count in self.counter.items():
                self.attrib[stat_name] = str(count)

        return super().append(subelement)


class ElementProperties(Element):
    """<properties> element"""

    def __init__(self):
        super().__init__("properties")

    def add_property(self, name: str, value: str):
        SubElement(self, "property", {"name": name, "value": value})


class ElementTestCase(Element):
    """<testcase> element = a single diagnosis"""

    def __init__(self, diagnosis: DiagnosisModel):
        # create <testcase> element
        attrib = {
            "name": diagnosis.loc_path,
            "classname": diagnosis.code,
            "line": str(diagnosis.line),
        }

        if diagnosis.extras.file and not diagnosis.extras.file.is_stdin:
            attrib["file"] = diagnosis.extras.file.filepath

        super().__init__("testcase", attrib)

        # determine result type
        match diagnosis.type:
            case "error":
                result_name = "error"
                stat_name = "errors"
            case "failure":
                result_name = "failure"
                stat_name = "failures"
            case "warning":
                result_name = "skipped"
                stat_name = "skipped"

        self.stat_name = stat_name
        self.attrib[stat_name] = "1"

        # create result element
        result_element = SubElement(self, result_name, {"message": diagnosis.summary})
        result_element.text = diagnosis.msg



from __future__ import annotations

import io
import os
import textwrap
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import (
    BaseModel,
    Field,
    StringConstraints,
    field_validator,
    model_validator,
)

if TYPE_CHECKING:
    from typing import Self


class DiagnosisModel(BaseModel):
    """A diagnosis reported by the analyzer."""

    # this class is a data model that conveys the same information as `tugboat.types.Diagnosis`
    # but is implemented using `pydantic.BaseModel`

    line: Annotated[
        int,
        Field(
            default=0,
            description=(
                "1-based line number in the YAML input where the issue begins. "
                "Line numbers continue across documents within the same multi-document stream."
            ),
        ),
    ]

    column: Annotated[
        int,
        Field(
            default=0,
            description=(
                "1-based column number for the first problematic character on that line."
            ),
        ),
    ]

    type: Annotated[
        Literal["error", "failure", "warning"],
        Field(
            default="failure",
            description=(
                "Diagnostic severity reported by the analyzer.\n"
                "* `error`: analysis stopped because processing could not continue.\n"
                "* `failure`: definite rule violation that prevents success.\n"
                "* `warning`: potential issue that should be reviewed.\n"
            ),
        ),
    ]

    code: Annotated[
        str,
        StringConstraints(
            strip_whitespace=True,
            to_upper=True,
            pattern=r"^[a-zA-Z]+\d+$",
        ),
        Field(
            description=(
                "Rule identifier for this diagnosis. "
                "Look up details at https://argo-tugboat.readthedocs.io/en/stable/search.html?q=CODE by replacing `CODE` with the value provided here."
            )
        ),
    ]

    loc: Annotated[
        tuple[str | int, ...],
        Field(
            description=(
                "JSONPath-like list that walks from the manifest root to the problematic field."
                'For example, an issue in `spec.containers[0].name` becomes `["spec", "containers", 0, "name"]`.'
            )
        ),
    ]

    summary: Annotated[
        str,
        Field(
            default="",
            description="Short title for the diagnosis, most often the rule name.",
        ),
    ]

    msg: Annotated[
        str,
        Field(
            description="Full sentence that explains the issue and may include fix guidance.",
        ),
    ]

    input: Annotated[
        str | int | bool | float | Any | None,
        Field(
            default=None,
            description="Original manifest value that triggered the diagnosis, when available.",
        ),
    ]

    fix: Annotated[
        str | None,
        Field(
            default=None,
            description=(
                "Analyzer's suggested fix presented as plain text. Treat this as guidance, not a guaranteed solution."
            ),
        ),
    ]

    extras: Annotated[
        Extras,
        Field(
            default_factory=lambda: Extras.model_validate({}),
            description="Structured container for any additional context (see `Extras`).",
        ),
    ]

    @field_validator("msg")
    @classmethod
    def _validate_msg_(cls, v: str) -> str:
        return textwrap.dedent(v).strip()

    @model_validator(mode="after")
    def _post_init_(self) -> Self:
        # create summary if missing
        if not self.summary:
            first_sentence, *_ = self.msg.split("\n", 1)
            first_sentence, *_ = first_sentence.split(". ", 1)
            self.summary = first_sentence.rstrip(" .")

        return self

    @property
    def loc_path(self) -> str:
        """Return a JSONPath-like string representation of the `loc` field."""
        with io.StringIO() as buf:
            for part in self.loc:
                if isinstance(part, int):
                    buf.write(f"[{part}]")
                else:
                    buf.write(f".{part}")

            if path := buf.getvalue():
                return path

        return "."


class Extras(BaseModel):
    """Additional context supplied alongside a diagnosis."""

    file: Annotated[
        FilesystemMetadata | None,
        Field(
            default=None,
            description="Filesystem-level context (file path) for the manifest related to the diagnosis.",
        ),
    ]

    helm: Annotated[
        HelmMetadata | None,
        Field(
            default=None,
            description="Helm metadata (chart name and template path) for the manifest. Present only when Helm rendered the manifest.",
        ),
    ]

    manifest: Annotated[
        ManifestMetadata | None,
        Field(
            default=None,
            description="Manifest-level context (API version, name, file path) related to the diagnosis.",
        ),
    ]


class FilesystemMetadata(BaseModel):
    """Metadata describing the filesystem that supplied the manifest."""

    filepath: Annotated[
        str,
        Field(
            description="Filesystem path that supplied the manifest.",
        ),
    ]

    @property
    def is_stdin(self) -> bool:
        if self.filepath == "<stdin>":
            return True
        if os.path.realpath(self.filepath) == os.path.realpath("/dev/stdin"):
            return True
        return False


class HelmMetadata(BaseModel):
    """Metadata describing the Helm chart that produced the manifest."""

    chart: Annotated[
        str,
        Field(
            description="Name of the Helm chart that rendered this manifest.",
            examples=["my-chart"],
        ),
    ]

    template: Annotated[
        str,
        Field(
            description="Relative path in the chart to the template that emitted the manifest.",
            examples=[
                "templates/workflow.yaml",
            ],
        ),
    ]


class ManifestMetadata(BaseModel):
    """Metadata describing the manifest that produced the diagnosis."""

    group: Annotated[
        str,
        Field(
            description="Kubernetes API group (e.g., `argoproj.io` or ``). The empty string represents the core API group.",
        ),
    ]

    kind: Annotated[
        str,
        Field(
            description="Kubernetes kind.",
            examples=["Pod", "Workflow"],
        ),
    ]

    name: Annotated[
        str | None,
        Field(
            default=None,
            description=(
                "Resource identifier taken from manifest metadata (`metadata.name` or `metadata.generateName`) to help locate the manifest."
            ),
        ),
    ]

    @property
    def fqk(self) -> str:
        """Return the fully qualified kind."""
        kind = self.kind.lower()
        if self.group:
            return f"{kind}.{self.group}"
        return kind

    @property
    def fqkn(self) -> str:
        """Return the fully qualified kind and name."""
        if not self.name:
            raise ValueError("name is not set")
        return f"{self.fqk}/{self.name}"


DiagnosisModel.model_rebuild()  # required for subclassing



"""
:py:mod:`tugboat.schemas` contains classes representing the schemas of Argo Workflows.

All the listed classes are :py:class:`pydantic.BaseModel` classes used for input
data validation. These classes are immutable and do not allow any extra fields.
This ensures that the input data strictly follows the schema and that the data
model remains consistent throughout the program's execution, even when data is
transferred between different modules.

All the classes are derived from Argo Workflows' official `Field Reference`_
documentation. If you find any discrepancies between the classes and the official
documentation, please report it as a bug.

Currently, many of the members are typed as :py:data:`typing.Any`.
This means that the field is not yet used in the codebase but is included in the
schema for future compatibility. This also provides the basic ability to validate
if there are any incompatible fields in the input data.

.. _Field Reference: https://argo-workflows.readthedocs.io/en/latest/fields/
"""

__all__ = [
    "Arguments",
    "Artifact",
    "ContainerNode",
    "ContainerTemplate",
    "CronWorkflow",
    "DagTask",
    "Manifest",
    "Parameter",
    "ScriptTemplate",
    "Step",
    "Template",
    "Workflow",
    "WorkflowTemplate",
]

from tugboat.core import hookimpl
from tugboat.schemas.arguments import Arguments, Artifact, Parameter
from tugboat.schemas.cron_workflow import CronWorkflow
from tugboat.schemas.debug import DebugManifest
from tugboat.schemas.manifest import Manifest
from tugboat.schemas.template import (
    ContainerNode,
    ContainerTemplate,
    DagTask,
    ScriptTemplate,
    Step,
    Template,
)
from tugboat.schemas.workflow import Workflow, WorkflowTemplate


@hookimpl
def parse_manifest(manifest: dict):
    """
    Parse the manifest and return the corresponding data model.

    :meta private:
    """
    match manifest.get("apiVersion"):
        case "argoproj.io/v1alpha1":
            match manifest.get("kind"):
                case "CronWorkflow":
                    return CronWorkflow.model_validate(manifest)
                case "Workflow":
                    return Workflow.model_validate(manifest)
                case "WorkflowTemplate":
                    return WorkflowTemplate.model_validate(manifest)

        case "tugboat.example.com/v1":
            match manifest.get("kind"):
                case "Debug":
                    return DebugManifest.model_validate(manifest)



from __future__ import annotations

import functools

import pluggy

hookimpl = pluggy.HookimplMarker("tugboat")
"""The hook implementation marker for the Tugboat framework."""


@functools.lru_cache(1)
def get_plugin_manager() -> pluggy.PluginManager:
    pm = pluggy.PluginManager("tugboat")
    pm.load_setuptools_entrypoints("tugboat")

    # hook specs
    import tugboat.hookspecs.core
    import tugboat.hookspecs.workflow

    pm.add_hookspecs(tugboat.hookspecs.core)
    pm.add_hookspecs(tugboat.hookspecs.workflow)

    # built-in hook implementations
    import tugboat.analyzers.container
    import tugboat.analyzers.cron_workflow
    import tugboat.analyzers.step
    import tugboat.analyzers.template
    import tugboat.analyzers.template.inputs
    import tugboat.analyzers.template.outputs
    import tugboat.analyzers.workflow
    import tugboat.analyzers.workflow_template
    import tugboat.schemas

    pm.register(tugboat.analyzers.container)
    pm.register(tugboat.analyzers.cron_workflow)
    pm.register(tugboat.analyzers.step)
    pm.register(tugboat.analyzers.template)
    pm.register(tugboat.analyzers.template.inputs)
    pm.register(tugboat.analyzers.template.outputs)
    pm.register(tugboat.analyzers.workflow)
    pm.register(tugboat.analyzers.workflow_template)
    pm.register(tugboat.schemas)

    return pm



#! /user/bin/env python3
"""
Checks how many colors are supported by the terminal.

Acknowledgements
================
The logic of this script is primarily based on the implementation found in:
- https://github.com/nodejs/node/blob/v24.2.0/lib/internal/tty.js
- https://github.com/chalk/supports-color
"""
from __future__ import annotations

__all__ = ["ColorSupport", "check_supported_colors"]

import enum
import os
import re
import sys
import typing

if typing.TYPE_CHECKING:
    from typing import TextIO


class ColorSupport(enum.Enum):
    """
    Enum for color support.
    """

    NONE = enum.auto()
    """No color support."""

    BASIC = enum.auto()
    """Basic color support (16 colors)."""

    ANSI_256 = enum.auto()
    """256 color support."""

    TRUE_COLOR = enum.auto()
    """True color support (24-bit, 16 million colors)."""


def check_supported_colors(stream: TextIO | None = None) -> ColorSupport:
    """
    Check how many colors are supported by the terminal.
    """
    # User specified behaviors
    match os.getenv("FORCE_COLOR"):  # https://force-color.org/
        case "" | "1" | "true":
            return ColorSupport.BASIC
        case "2":
            return ColorSupport.ANSI_256
        case "3":
            return ColorSupport.TRUE_COLOR

    if "NO_COLOR" in os.environ:  # https://no-color.org/
        return ColorSupport.NONE

    # dumb terminals
    if os.getenv("TERM") == "dumb":
        return ColorSupport.NONE

    # Azure DevOps
    if "TF_BUILD" in os.environ and "AGENT_NAME" in os.environ:
        return ColorSupport.BASIC

    # Check if the stream is a TTY
    if stream and not stream.isatty():
        return ColorSupport.NONE

    # Windows
    if sys.platform == "win32":
        if _has_env("ANSICON"):  # ANSICON
            return ColorSupport.TRUE_COLOR
        if _has_env("ConEmuANSI"):  # ConEmu
            return ColorSupport.ANSI_256

        version_info = sys.getwindowsversion()
        if version_info.major >= 10:  # Windows 10
            if version_info.build >= 14931:
                return ColorSupport.TRUE_COLOR
            if version_info.build >= 10586:
                return ColorSupport.ANSI_256

        return ColorSupport.BASIC

    # CI environments
    if _has_env("CI"):
        if _has_env("CIRCLECI", "GITEA_ACTIONS", "GITHUB_ACTIONS"):
            return ColorSupport.TRUE_COLOR
        if _has_env("APPVEYOR", "BUILDKITE", "DRONE", "GITLAB_CI", "TRAVIS"):
            return ColorSupport.ANSI_256

    # JetBrains TeamCity
    if _has_env("TEAMCITY_VERSION"):
        version_info = _split_version(os.getenv("TEAMCITY_VERSION", "0"))
        if version_info >= (9,):
            return ColorSupport.BASIC
        return ColorSupport.NONE

    # terminal emulators
    match os.getenv("TERM_PROGRAM"):
        case "Apple_Terminal":
            return ColorSupport.ANSI_256
        case "HyperTerm":
            return ColorSupport.TRUE_COLOR
        case "iTerm.app":
            version_info = _split_version(os.getenv("TERM_PROGRAM_VERSION", "0"))
            if version_info >= (3,):
                return ColorSupport.TRUE_COLOR
            return ColorSupport.ANSI_256
        case "MacTerm":
            return ColorSupport.TRUE_COLOR
        case "vscode":
            return ColorSupport.ANSI_256

    # $COLORTERM environment variable
    if os.getenv("COLORTERM") in ("truecolor", "24bit"):
        return ColorSupport.TRUE_COLOR

    # $TERM environment variable
    if term := os.getenv("TERM"):
        match term:
            case (
                "cons25"
                | "console"
                | "cygwin"
                | "dtterm"
                | "eterm"
                | "gnome"
                | "hurd"
                | "jfbterm"
                | "konsole"
                | "kterm"
                | "mlterm"
                | "putty"
                | "st"
            ):
                return ColorSupport.BASIC

            case "mosh" | "rxvt-unicode-24bit" | "terminator" | "xterm-kitty":
                return ColorSupport.TRUE_COLOR

        if any(keyword in term for keyword in ("truecolor", "24bit")):
            return ColorSupport.TRUE_COLOR

        term = term.lower()

        if term.endswith(("-256", "-256color")):
            return ColorSupport.ANSI_256

        if term.startswith(("rxvt", "screen", "vt100", "vt220", "xterm")):
            return ColorSupport.BASIC

        if any(
            keyword in term
            for keyword in ("ansi", "color", "cygwin", "direct", "linux")
        ):
            return ColorSupport.BASIC

        if re.match(r"con[0-9]*x[0-9]", term):
            return ColorSupport.BASIC

    if _has_env("COLORTERM"):
        return ColorSupport.BASIC

    # can't determine color support
    return ColorSupport.NONE


def _has_env(*names: str) -> bool:
    """Check if any of the given environment variable names exist."""
    return any(map(os.getenv, names))


def _split_version(v: str) -> tuple[int, ...]:
    """Split a version string into a tuple of integers."""
    return tuple(map(int, v.split(".")))


if __name__ == "__main__":
    print("Supported colors for stdout:", check_supported_colors(sys.stdout))
    print("Supported colors for stdout:", check_supported_colors(sys.stderr))



from __future__ import annotations

import itertools
import typing

from rapidfuzz.process import extractOne

from tugboat.analyzers.kubernetes import check_resource_name
from tugboat.analyzers.metrics import check_prometheus
from tugboat.constraints import accept_none, require_exactly_one, require_non_empty
from tugboat.core import get_plugin_manager, hookimpl
from tugboat.references import get_workflow_context
from tugboat.utils import (
    find_duplicate_names,
    join_with_and,
    prepend_loc,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.schemas import Workflow, WorkflowTemplate
    from tugboat.types import Diagnosis

    type WorkflowCompatible = Workflow | WorkflowTemplate


@hookimpl
def analyze(manifest: WorkflowCompatible) -> Iterator[Diagnosis]:
    # early escape if the manifest is not recognized
    if manifest.apiVersion != "argoproj.io/v1alpha1":
        return
    if manifest.kind not in ("Workflow", "WorkflowTemplate"):
        return

    # invoke checks
    pm = get_plugin_manager()

    if manifest.kind == "Workflow":
        manifest_diagnoses_generators = pm.hook.analyze_workflow(workflow=manifest)
    else:
        manifest_diagnoses_generators = pm.hook.analyze_workflow_template(
            workflow_template=manifest
        )
    yield from itertools.chain.from_iterable(manifest_diagnoses_generators)

    for idx_tmpl, template in enumerate(manifest.spec.templates or []):
        yield from prepend_loc.from_iterables(
            ["spec", "templates", idx_tmpl],
            pm.hook.analyze_template(template=template, workflow=manifest),
        )


# ----------------------------------------------------------------------------
# Workflow analyzers
# ----------------------------------------------------------------------------


@hookimpl(specname="analyze_workflow")
def check_metadata(workflow: Workflow) -> Iterator[Diagnosis]:
    yield from require_exactly_one(
        model=workflow.metadata,
        loc=("metadata",),
        fields=["name", "generateName"],
    )

    if workflow.metadata.name:
        yield from prepend_loc(
            ["metadata", "name"], check_resource_name(workflow.metadata.name)
        )

    if workflow.metadata.generateName:
        yield from prepend_loc(
            ["metadata", "generateName"],
            check_resource_name(workflow.metadata.generateName, is_generate_name=True),
        )


@hookimpl(specname="analyze_workflow")
def check_spec(workflow: Workflow) -> Iterator[Diagnosis]:
    yield from require_exactly_one(
        model=workflow.spec,
        loc=("spec",),
        fields=["templates", "workflowTemplateRef"],
    )

    if not workflow.spec.workflowTemplateRef:
        yield from require_non_empty(
            model=workflow.spec,
            loc=("spec",),
            fields=["entrypoint"],
        )


@hookimpl(specname="analyze_workflow")
def check_entrypoint(workflow: WorkflowCompatible) -> Iterator[Diagnosis]:
    if not workflow.spec.templates:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(workflow.spec.templates):
        yield {
            "code": "TPL101",
            "loc": ("spec", "templates", idx, "name"),
            "summary": "Duplicate template name",
            "msg": f"Template name '{name}' is duplicated.",
            "input": name,
        }

    # if the workflow has an entrypoint, check if it exists
    if (
        True
        and workflow.spec.entrypoint
        and workflow.spec.entrypoint not in workflow.template_dict
    ):
        entrypoints = sorted(workflow.template_dict)
        suggestion = None
        if result := extractOne(workflow.spec.entrypoint, entrypoints):
            suggestion, _, _ = result
        yield {
            "code": "WF201",
            "loc": ("spec", "entrypoint"),
            "summary": "Invalid entrypoint",
            "msg": f"""
                Entrypoint '{workflow.spec.entrypoint}' is not defined in any template.
                Defined entrypoints: {join_with_and(entrypoints)}
                """,
            "input": workflow.spec.entrypoint,
            "fix": suggestion,
        }


@hookimpl(specname="analyze_workflow")
def check_argument_parameters(workflow: WorkflowCompatible) -> Iterator[Diagnosis]:
    if not workflow.spec.arguments:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(workflow.spec.arguments.parameters or ()):
        yield {
            "code": "WF101",
            "loc": ("spec", "arguments", "parameters", idx, "name"),
            "summary": "Duplicate parameter name",
            "msg": f"Parameter name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each parameter
    for idx, param in enumerate(workflow.spec.arguments.parameters or []):
        loc = ("spec", "arguments", "parameters", idx)

        yield from require_non_empty(
            model=param,
            loc=loc,
            fields=["name"],
        )
        yield from accept_none(
            model=param,
            loc=loc,
            fields=["default", "enum"],
        )

        if workflow.kind == "Workflow":
            yield from require_exactly_one(
                model=param,
                loc=loc,
                fields=["value", "valueFrom"],
            )

        if param.valueFrom:
            yield from require_exactly_one(
                model=param.valueFrom,
                loc=(*loc, "valueFrom"),
                fields=[
                    "configMapKeyRef",
                    "expression",
                ],
            )
            yield from accept_none(
                model=param.valueFrom,
                loc=(*loc, "valueFrom"),
                fields=[
                    "default",
                    "event",
                    "jqFilter",
                    "jsonPath",
                    "parameter",
                    "path",
                    "supplied",
                ],
            )


@hookimpl(specname="analyze_workflow")
def check_argument_artifacts(workflow: WorkflowCompatible) -> Iterator[Diagnosis]:
    if not workflow.spec.arguments:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(workflow.spec.arguments.artifacts or ()):
        yield {
            "code": "WF102",
            "loc": ("spec", "arguments", "artifacts", idx, "name"),
            "summary": "Duplicate artifact name",
            "msg": f"Artifact name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each artifact; also count the number of times each name appears
    for idx, artifact in enumerate(workflow.spec.arguments.artifacts or []):
        loc = ("spec", "arguments", "artifacts", idx)

        yield from require_exactly_one(
            model=artifact,
            loc=loc,
            fields=[
                "artifactory",
                "azure",
                "gcs",
                "git",
                "hdfs",
                "http",
                "oss",
                "raw",
                "s3",
            ],
        )
        yield from accept_none(
            model=artifact,
            loc=loc,
            fields=[
                "archive",
                "archiveLogs",
                "artifactGC",
                "deleted",
                "from_",
                "fromExpression",
                "mode",
                "path",
                "recurseMode",
            ],
        )


@hookimpl(specname="analyze_workflow")
def check_metrics(workflow: WorkflowCompatible) -> Iterator[Diagnosis]:
    if not workflow.spec.metrics:
        return

    # build context
    # some additional variables should be available in the context
    ctx = get_workflow_context(workflow)
    ctx.parameters |= {
        ("workflow", "duration"),
        ("workflow", "status"),
    }

    # check metrics
    for idx, prom in enumerate(workflow.spec.metrics.prometheus or ()):
        for diagnosis in prepend_loc(
            ("spec", "metrics", "prometheus", idx), check_prometheus(prom, ctx)
        ):
            match diagnosis["code"]:
                case "internal:invalid-metric-name":
                    diagnosis["code"] = "WF301"
                case "internal:invalid-metric-label-name":
                    diagnosis["code"] = "WF302"
                case "internal:invalid-metric-label-value":
                    diagnosis["code"] = "WF303"
            yield diagnosis



from __future__ import annotations

import logging
import re
import typing

from rapidfuzz.process import extractOne

from tugboat.constraints import (
    accept_none,
    mutually_exclusive,
    require_exactly_one,
    require_non_empty,
)
from tugboat.core import get_plugin_manager, hookimpl
from tugboat.references import get_step_context
from tugboat.schemas import Arguments
from tugboat.types import Field
from tugboat.utils import (
    check_model_fields_references,
    check_value_references,
    find_duplicate_names,
    join_with_and,
    join_with_or,
    prepend_loc,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.references import Context
    from tugboat.schemas import (
        Artifact,
        Parameter,
        Step,
        Template,
        Workflow,
        WorkflowTemplate,
    )
    from tugboat.types import Diagnosis

    type WorkflowCompatible = Workflow | WorkflowTemplate

logger = logging.getLogger(__name__)


@hookimpl
def analyze_step(
    step: Step, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    yield from require_exactly_one(
        model=step,
        loc=(),
        fields=["template", "templateRef", "inline"],
    )
    yield from mutually_exclusive(
        model=step,
        loc=(),
        fields=["withItems", "withParam", "withSequence"],
    )

    if step.onExit:
        yield {
            "code": "STP901",
            "loc": ("onExit",),
            "summary": "Deprecated field",
            "msg": "Field 'onExit' is deprecated. Please use 'hooks[exit].template' instead.",
            "input": Field("onExit"),
        }


@hookimpl(specname="analyze_step")
def check_argument_parameters(
    step: Step, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if not step.arguments:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(step.arguments.parameters or ()):
        yield {
            "code": "STP102",
            "loc": ("arguments", "parameters", idx, "name"),
            "summary": "Duplicate parameter name",
            "msg": f"Parameter name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each parameter
    ctx = get_step_context(workflow, template, step)

    for idx, param in enumerate(step.arguments.parameters or ()):
        yield from prepend_loc(
            ("arguments", "parameters", idx),
            _check_argument_parameter_fields(param, ctx),
        )


def _check_argument_parameter_fields(
    param: Parameter, context: Context
) -> Iterable[Diagnosis]:
    yield from require_non_empty(
        model=param,
        loc=(),
        fields=["name"],
    )
    yield from mutually_exclusive(
        model=param,
        loc=(),
        fields=["value", "valueFrom"],
    )

    if param.valueFrom:
        yield from require_exactly_one(
            model=param.valueFrom,
            loc=("valueFrom",),
            fields=[
                "configMapKeyRef",
                "expression",
                "parameter",
            ],
        )
        yield from accept_none(
            model=param.valueFrom,
            loc=("valueFrom",),
            fields=[
                "event",
                "globalName",
                "jqFilter",
                "jsonPath",
                "path",
                "supplied",
            ],
        )

    for diag in check_model_fields_references(param, context.parameters):
        match diag["code"]:
            case "VAR002":
                diag["code"] = "STP301"
                if metadata := diag.get("ctx", {}).get("reference"):
                    ref = metadata["found:str"]
                    diag["msg"] = (
                        f"The parameter reference '{ref}' used in parameter '{param.name}' is invalid."
                    )
        yield diag


@hookimpl(specname="analyze_step")
def check_argument_parameters_usage(
    step: Step, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    # early exit: referenced template not found
    ref_template = _get_template_by_ref(step, workflow)
    if not ref_template:
        return

    # prepare arguments in case not provided
    if step.arguments:
        arguments = step.arguments
    else:
        arguments = Arguments()

    # check for redundant parameters
    if ref_template.inputs:
        expected_params = set(ref_template.inputs.parameter_dict)
    else:
        expected_params = set()

    for idx, param in enumerate(arguments.parameters or ()):
        if param.name and param.name not in expected_params:
            suggestion = None
            if result := extractOne(param.name, expected_params):
                suggestion, _, _ = result

            yield {
                "type": "warning",  # redundant parameter does not break the workflow
                "code": "STP304",
                "loc": ("arguments", "parameters", idx, "name"),
                "summary": "Unexpected parameter",
                "msg": f"Parameter '{param.name}' is not expected by the template '{ref_template.name}'.",
                "input": param.name,
                "fix": suggestion,
            }

    # check for missing parameters
    if ref_template.inputs:
        required_params = set()
        for name, model in ref_template.inputs.parameter_dict.items():
            if model.default is None and model.value is None:
                required_params.add(name)

        missing_parameters = required_params.difference(arguments.parameter_dict)
        if missing_parameters:
            yield {
                "code": "STP305",
                "loc": ("arguments", "parameters"),
                "summary": "Missing parameters",
                "msg": (
                    f"Parameters {join_with_and(missing_parameters)} are required by the template '{ref_template.name}' but are not provided."
                ),
                "input": Field("parameters"),
            }


def _get_template_by_ref(step: Step, workflow: WorkflowCompatible) -> Template | None:
    if step.template:
        return workflow.template_dict.get(step.template)
    if step.templateRef and step.templateRef.name == workflow.metadata.name:
        return workflow.template_dict.get(step.templateRef.template)
    return None


@hookimpl(specname="analyze_step")
def check_argument_artifacts(
    step: Step, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if not step.arguments:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(step.arguments.artifacts or ()):
        yield {
            "code": "STP103",
            "loc": ("arguments", "artifacts", idx, "name"),
            "summary": "Duplicate artifact name",
            "msg": f"Artifact name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each parameter
    ctx = get_step_context(workflow, template, step)

    for idx, artifact in enumerate(step.arguments.artifacts or []):
        yield from prepend_loc(
            ("arguments", "artifacts", idx),
            _check_argument_artifact_fields(artifact, ctx),
        )


def _check_argument_artifact_fields(
    artifact: Artifact, context: Context
) -> Iterable[Diagnosis]:
    yield from require_non_empty(
        model=artifact,
        loc=(),
        fields=["name"],
    )
    yield from mutually_exclusive(
        model=artifact,
        loc=(),
        fields=[
            "artifactory",
            "azure",
            "from_",
            "fromExpression",
            "gcs",
            "git",
            "hdfs",
            "http",
            "oss",
            "raw",
            "s3",
        ],
    )
    yield from accept_none(
        model=artifact,
        loc=(),
        fields=[
            "archive",
            "archiveLogs",
            "artifactGC",
            "deleted",
            "globalName",
        ],
    )

    if artifact.from_:
        # `from` value can be either wrapped by the cruely brackets or not
        from_ = artifact.from_.strip()

        # when it is unwrapped, it should be a reference to an artifact
        # EXAMPLE> artifact: inputs.artifacts.artifact-1
        if re.fullmatch(r"[a-zA-Z0-9.-]+", from_):
            ref = tuple(from_.split("."))
            if ref not in context.artifacts:
                closest = context.artifacts.find_closest(ref)
                yield {
                    "code": "STP302",
                    "loc": ("from",),
                    "summary": "Invalid reference",
                    "msg": f"The reference '{from_}' used in artifact '{artifact.name}' is invalid.",
                    "input": artifact.from_,
                    "fix": ".".join(closest),
                }

        # when it is wrapped, it may be a reference to a parameter or an artifact
        mixed_references = context.parameters + context.artifacts
        for diag in prepend_loc(
            ("from",), check_value_references(from_, mixed_references)
        ):
            match diag["code"]:
                case "VAR002":
                    diag["code"] = "STP302"
                    if metadata := diag.get("ctx", {}).get("reference"):
                        ref = metadata["found:str"]
                        diag["msg"] = (
                            f"The reference '{ref}' used in artifact '{artifact.name}' is invalid."
                        )
            yield diag

    # TODO fromExpression

    if artifact.raw:
        for diag in prepend_loc(
            ("raw", "data"),
            check_value_references(artifact.raw.data, context.parameters),
        ):
            match diag["code"]:
                case "VAR002":
                    diag["code"] = "STP303"
                    if metadata := diag.get("ctx", {}).get("reference"):
                        ref = metadata["found:str"]
                        diag["msg"] = (
                            f"""
                            The parameter reference '{ref}' used in artifact '{artifact.name}' is invalid.
                            Note: Only parameter references are allowed here, even though this is an artifact object.
                            """
                        )
            yield diag


@hookimpl(specname="analyze_step")
def check_argument_artifact_usage(
    step: Step, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    # early exit: referenced template not found
    ref_template = _get_template_by_ref(step, workflow)
    if not ref_template:
        return

    # prepare arguments in case not provided
    if step.arguments:
        arguments = step.arguments
    else:
        arguments = Arguments()

    # check for redundant artifacts
    if ref_template.inputs:
        expected_artifacts = set(ref_template.inputs.artifact_dict)
    else:
        expected_artifacts = set()

    for idx, artifact in enumerate(arguments.artifacts or ()):
        if artifact.name and artifact.name not in expected_artifacts:
            suggestion = None
            if result := extractOne(artifact.name, expected_artifacts):
                suggestion, _, _ = result

            yield {
                "type": "warning",  # redundant artifact does not break the workflow
                "code": "STP306",
                "loc": ("arguments", "artifacts", idx, "name"),
                "summary": "Unexpected artifact",
                "msg": f"Artifact '{artifact.name}' is not expected by the template '{ref_template.name}'.",
                "input": artifact.name,
                "fix": suggestion,
            }

    # check for missing artifacts
    if ref_template.inputs:
        required_artifacts = set()
        for name, model in ref_template.inputs.artifact_dict.items():
            if (
                False
                # -- optional artifact --
                or model.optional
                # -- defined in template --
                or model.artifactory
                or model.azure
                or model.gcs
                or model.git
                or model.hdfs
                or model.http
                or model.oss
                or model.raw
                or model.s3
            ):
                continue
            required_artifacts.add(name)

        missing_artifacts = required_artifacts.difference(arguments.artifact_dict)
        if missing_artifacts:
            yield {
                "code": "STP307",
                "loc": ("arguments", "artifacts"),
                "summary": "Missing artifacts",
                "msg": (
                    f"Artifacts {join_with_and(missing_artifacts)} are required by the template '{ref_template.name}' but are not provided."
                ),
                "input": Field("artifacts"),
            }


@hookimpl(specname="analyze_step")
def check_referenced_template(
    step: Step, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if step.template:
        yield from prepend_loc(
            ("template",),
            _check_referenced_template(step.template, template, workflow),
        )

    elif step.templateRef:
        if step.templateRef.name == workflow.metadata.name:
            yield from prepend_loc(
                ("templateRef", "template"),
                _check_referenced_template(
                    step.templateRef.template, template, workflow
                ),
            )
        else:
            logger.debug(
                "Step '%s': Referenced template '%s' is not the same as current workflow '%s'. Skipping.",
                step.name,
                step.templateRef.name,
                workflow.name,
            )


def _check_referenced_template(
    target_template_name: str, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if target_template_name == template.name:
        yield {
            "type": "warning",
            "code": "STP201",
            "loc": (),
            "summary": "Self-referencing",
            "msg": "Self-referencing may cause infinite recursion.",
            "input": target_template_name,
        }

    if target_template_name not in workflow.template_dict:
        templates = set(workflow.template_dict)
        templates -= {template.name}

        suggestion = None
        if result := extractOne(target_template_name, templates):
            suggestion, _, _ = result

        yield {
            "code": "STP202",
            "loc": (),
            "summary": "Template not found",
            "msg": f"""
                Template '{target_template_name}' does not exist in the workflow.
                Available templates: {join_with_or(templates)}
                """,
            "input": target_template_name,
            "fix": suggestion,
        }


@hookimpl(specname="analyze_step")
def check_fields_references(
    step: Step, template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    ctx = get_step_context(workflow, template, step)
    yield from check_model_fields_references(
        step,
        ctx.parameters,
        exclude=["arguments", "inline", "withItems", "withSequence"],
    )


@hookimpl(specname="analyze_step")
def check_inline_template(
    step: Step, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if not step.inline:
        return

    # check inline template type
    for diagnosis in accept_none(
        model=step.inline,
        loc=("inline",),
        fields=("dag", "steps"),
    ):
        diagnosis["code"] = "STP401"
        diagnosis["summary"] = "Invalid inline template type"
        diagnosis["msg"] = (
            """
            Nested steps or DAGs will result in an invalid step.
            Please use simple steps or a DAG at the top level of the template.
            """
        )
        yield diagnosis

    # check inline template definitions
    pm = get_plugin_manager()

    for diagnosis in prepend_loc.from_iterables(
        ("inline",),
        pm.hook.analyze_template(template=step.inline, workflow=workflow),
    ):
        if diagnosis["loc"] == ("inline", "name"):
            # template name is not required
            continue
        if diagnosis["loc"][:2] in {("inline", "steps"), ("inline", "dag")}:
            # step templates or dag templates are not allowed
            continue

        yield diagnosis



from __future__ import annotations

__all__ = [
    "ContainerNode",
    "ContainerTemplate",
    "ScriptTemplate",
    "Step",
    "Template",
]

import functools
import itertools
import os
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.arguments import Arguments
from tugboat.schemas.basic import Array, Dict
from tugboat.schemas.metrics import Metrics
from tugboat.schemas.template.container import (
    ContainerNode,
    ContainerSetTemplate,
    ContainerTemplate,
    ScriptTemplate,
)

if TYPE_CHECKING:
    from typing import Literal

    type TemplateType = Literal[
        "container",
        "containerSet",
        "dag",
        "data",
        "http",
        "resource",
        "script",
        "steps",
        "suspend",
    ]

if os.getenv("DOCUTILSCONFIG"):
    __all__ = [
        "DagTemplate",
        "SuspendTemplate",
        "TemplateRef",
    ]


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class Template(_BaseModel):
    """
    `Template`_ is a reusable and composable unit of execution in a workflow.

    .. _Template: https://argo-workflows.readthedocs.io/en/latest/fields/#template
    """

    activeDeadlineSeconds: int | str | None = None
    annotations: Dict[str, str] | None = None
    automountServiceAccountToken: bool | None = None
    container: ContainerTemplate | None = None
    containerSet: ContainerSetTemplate | None = None
    daemon: bool | None = None
    dag: DagTemplate | None = None
    failFast: bool | None = None
    inputs: Arguments | None = None
    metrics: Metrics | None = None
    name: str | None = None
    nodeSelector: Dict[str, str] | None = None
    outputs: Arguments | None = None
    parallelism: int | None = None
    podSpecPatch: str | None = None
    priority: int | None = None
    priorityClassName: str | None = None
    schedulerName: str | None = None
    script: ScriptTemplate | None = None
    serviceAccountName: str | None = None
    steps: Array[Array[Step]] | None = None
    suspend: SuspendTemplate | None = None
    timeout: str | None = None

    affinity: Any | None = None
    archiveLocation: Any | None = None
    data: Any | None = None
    executor: Any | None = None
    hostAliases: Array[Any] | None = None
    http: Any | None = None
    initContainers: Array[Any] | None = None
    memoize: Any | None = None
    metadata: Any | None = None
    plugins: Any | None = None
    resource: Any | None = None
    retryStrategy: Any | None = None
    securityContext: Any | None = None
    sidecars: Array[Any] | None = None
    synchronization: Any | None = None
    tolerations: Array[Any] | None = None
    volumes: Array[Any] | None = None

    def __hash__(self):
        return hash((self.name, self.container, self.script, self.steps))

    @functools.cached_property
    def type(self) -> TemplateType | None:
        """Return the template type as a string."""
        if self.container is not None:
            return "container"
        if self.containerSet is not None:
            return "containerSet"
        if self.dag is not None:
            return "dag"
        if self.data is not None:
            return "data"
        if self.http is not None:
            return "http"
        if self.resource is not None:
            return "resource"
        if self.script is not None:
            return "script"
        if self.steps is not None:
            return "steps"
        if self.suspend is not None:
            return "suspend"
        return None

    @functools.cached_property
    def step_dict(self) -> dict[str, Step]:
        """
        Step name to step data model mapping.

        .. note::

           Duplicate names will be overwritten and the empty name will be ignored.
        """
        return {
            step.name: step
            for step in itertools.chain.from_iterable(self.steps or ())
            if step.name
        }


# ----------------------------------------------------------------------------
# dags
# ----------------------------------------------------------------------------
class _StepBase(_BaseModel):

    arguments: Arguments | None = None
    inline: Template | None = None
    name: str
    onExit: str | None = None
    template: str | None = None
    templateRef: TemplateRef | None = None
    when: str | None = None
    withItems: Array[str | bool | int | dict[str, Any]] | None = None
    withParam: str | None = None

    continueOn: Any | None = None
    hooks: Any | None = None
    withSequence: Any | None = None

    def __hash__(self):
        return hash((self.name, self.template, self.templateRef))


class DagTask(_StepBase):
    """
    `DAGTask` represents a node in the graph during DAG execution.

    .. _DAGTask: https://argo-workflows.readthedocs.io/en/latest/fields/#dagtask
    """

    dependencies: Array[str] | None = None
    depends: str | None = None


class DagTemplate(_BaseModel):
    """
    `DAGTemplate`_ is a template subtype for directed acyclic graph templates.

    .. _DAGTemplate: https://argo-workflows.readthedocs.io/en/latest/fields/#dagtemplate
    """

    failFast: bool | None = None
    target: str | None = None
    tasks: Array[DagTask]


# ----------------------------------------------------------------------------
# steps
# ----------------------------------------------------------------------------
class Step(_StepBase):
    """
    `Step`_ is a reference to a template to execute in a series of step.

    .. _Step: https://argo-workflows.readthedocs.io/en/latest/fields/#workflowstep
    """


class TemplateRef(_BaseModel):
    """
    `TemplateRef`_ is a reference of template resource.

    .. _TemplateRef: https://argo-workflows.readthedocs.io/en/latest/fields/#templateref
    """

    clusterScope: bool | None = None
    name: str
    template: str


# ----------------------------------------------------------------------------
# suspend
# ----------------------------------------------------------------------------
class SuspendTemplate(_BaseModel):
    """
    `SuspendTemplate`_ is a template subtype to suspend a workflow at a predetermined point in time.

    .. _SuspendTemplate: https://argo-workflows.readthedocs.io/en/latest/fields/#suspendtemplate
    """

    duration: str | None = None



from __future__ import annotations

from collections.abc import MutableSet

from pydantic import BaseModel, Field, InstanceOf
from rapidfuzz.distance.DamerauLevenshtein import (
    distance as dameau_levenshtein_distance,
)
from rapidfuzz.distance.DamerauLevenshtein import (
    normalized_distance as dameau_levenshtein_normalized_distance,
)

type _TR = tuple[str, ...]
"""Type alias for reference, which is a tuple of strings."""


class _AnyStr(str):
    """An object that matches any string."""

    def __eq__(self, value):
        return isinstance(value, str)

    def __repr__(self):
        return "ANY"

    def __hash__(self):
        return super().__hash__()


AnyStr = _AnyStr(":any:")
"""
A special object that matches any string.
"""


class ReferenceCollection(MutableSet[_TR]):
    """
    A :py:class:`set`-like collection of references, handling the special logic
    of :py:data:`AnyStr`.
    """

    def __init__(self):
        self._static = set()
        self._dynamic = []

    def __iter__(self):
        yield from self._static
        yield from self._dynamic

    def __len__(self):
        return len(self._static) + len(self._dynamic)

    def __contains__(self, item):
        return item in self._static or item in self._dynamic

    def __repr__(self):
        return f'{{{", ".join(repr(item) for item in self)}}}'

    def __deepcopy__(self, memo):
        new = type(self)()
        new._static = self._static.copy()
        new._dynamic = self._dynamic.copy()
        return new

    def __add__(self, other):
        new = type(self)()
        new._static = self._static | other._static
        new._dynamic = self._dynamic + other._dynamic
        return new

    def add(self, value):
        if not isinstance(value, tuple | list):
            raise TypeError("value must be a tuple or a list")
        if any(isinstance(item, _AnyStr) for item in value):
            self._dynamic.append(value)
        else:
            self._static.add(value)

    def discard(self, value):
        raise NotImplementedError

    def find_closest(self, target: _TR) -> _TR:
        """
        Find the closest match for a given reference in a list of reference.
        """
        # NOTE this algorithm is heuristic

        # group the candidates by their distance to the target reference
        distance_grouped_candidates: dict[int, list[_TR]] = {}
        for candidate in self:
            dist = dameau_levenshtein_distance(target, candidate)
            distance_grouped_candidates.setdefault(dist, []).append(candidate)

        # find the closest group
        closest_candidates = []
        if distance_grouped_candidates:
            closest_distance = min(distance_grouped_candidates)
            closest_candidates = distance_grouped_candidates[closest_distance]

        # sort by the distance of each element
        def _calculate_distance(candidate: _TR) -> tuple[float, ...]:
            # calculate the normalized distance for each element
            base_distance = (
                dameau_levenshtein_normalized_distance(a, b)
                for a, b in zip(target, candidate, strict=False)
            )

            if len(target) != len(candidate):
                # if the lengths are different, add a penalty to the distance
                return (*base_distance, 2.0)
            else:
                return tuple(base_distance)

        closest_candidates = sorted(closest_candidates, key=_calculate_distance)

        # return the closest match
        for candidate in closest_candidates:
            # in case the candidate contains `AnyStr`, rebuild it with the target reference
            if any(item is AnyStr for item in candidate):
                if len(target) < len(candidate):
                    continue
                return tuple(
                    a if isinstance(b, _AnyStr) else b
                    for a, b in zip(target, candidate, strict=False)
                )
            return candidate

        return ()


class Context(BaseModel):
    """
    Data class for storing available references within a given scope.

    You can use the :py:keyword:`in` syntax to check if a reference exists in the context.
    For example:

    .. code-block:: python

       ctx = get_workflow_context(workflow)
       if ("inputs", "parameters", "message") in ctx.parameters:
           print("The variable `inputs.parameters.message` is available.")
    """

    parameters: InstanceOf[ReferenceCollection] = Field(
        default_factory=ReferenceCollection
    )
    """Collection of available parameters.
    """

    artifacts: InstanceOf[ReferenceCollection] = Field(
        default_factory=ReferenceCollection
    )
    """Collection of available artifacts.
    """



from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.basic import Array, KeyValuePair


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class Metrics(_BaseModel):
    """
    `Metrics`_ are a list of metrics emitted from a :py:class:`~tugboat.schemas.workflow.Workflow`
    or :py:class:`~tugboat.schemas.template.Template`.

    .. _Metrics: https://argo-workflows.readthedocs.io/en/latest/fields/#metrics
    """

    prometheus: Array[Prometheus]


class Prometheus(_BaseModel):
    """
    `Prometheus`_ is a prometheus metric to be emitted.

    .. _Prometheus: https://argo-workflows.readthedocs.io/en/latest/fields/#prometheus
    """

    counter: Counter | None = None
    gauge: Gauge | None = None
    help: str
    histogram: Histogram | None = None
    labels: Array[KeyValuePair] | None = None
    name: str
    when: str | None = None


class Counter(_BaseModel):
    """
    `Counter` is a `prometheus counter`_ metric.

    .. _Counter: https://argo-workflows.readthedocs.io/en/latest/fields/#counter
    .. _prometheus counter: https://prometheus.io/docs/concepts/metric_types/#counter
    """

    value: str


class Gauge(_BaseModel):
    """
    `Gauge`_ is a `prometheus gauge`_  metric.

    .. _Gauge: https://argo-workflows.readthedocs.io/en/latest/fields/#gauge
    .. _prometheus gauge: https://prometheus.io/docs/concepts/metric_types/#gauge
    """

    operation: str | None = None
    realtime: bool | None = None
    value: str


class Histogram(_BaseModel):
    """
    `Histogram`_ is a `prometheus histogram`_ metric.

    .. _Histogram: https://argo-workflows.readthedocs.io/en/latest/fields/#histogram
    .. _prometheus histogram: https://prometheus.io/docs/concepts/metric_types/#histogram
    """

    buckets: Array[int | float]
    value: str



from __future__ import annotations

import functools

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.arguments.artifact import Artifact
from tugboat.schemas.arguments.parameter import Parameter
from tugboat.schemas.basic import Array


class Arguments(BaseModel):
    """
    `Arguments` to a template.

    .. _Arguments: https://argo-workflows.readthedocs.io/en/latest/fields/#arguments
    """

    model_config = ConfigDict(extra="forbid", frozen=True)

    artifacts: Array[Artifact] | None = None
    """Artifacts is the list of artifacts to pass to the template or workflow."""
    parameters: Array[Parameter] | None = None
    """Parameters is the list of parameters to pass to the template or workflow."""

    @functools.cached_property
    def parameter_dict(self) -> dict[str, Parameter]:
        """
        Parameter name to :py:attr:`parameters` mapping.
        """
        return {
            parameter.name: parameter
            for parameter in self.parameters or ()
            if parameter.name
        }

    @functools.cached_property
    def artifact_dict(self) -> dict[str, Artifact]:
        """
        Artifact name to :py:attr:`artifacts` mapping.
        """
        return {
            artifact.name: artifact
            for artifact in self.artifacts or ()
            if artifact.name
        }



"""
As the engine hums, the boats remembers why it sails.
"""

from __future__ import annotations

__all__ = [
    "DiagnosisModel",
    "FilesystemMetadata",
    "HelmMetadata",
    "ManifestMetadata",
    "analyze_manifest",
    "analyze_yaml_document",
    "analyze_yaml_stream",
]

import hashlib
import io
import logging
import os
import re
import typing
from typing import cast

import ruamel.yaml
import ruamel.yaml.error
from ruamel.yaml.comments import CommentedBase, CommentedMap
from ruamel.yaml.tokens import CommentToken

from tugboat.engine.helpers import get_suppression_codes, translate_marked_yaml_error
from tugboat.engine.linecol import get_line_column
from tugboat.engine.mainfest import analyze_manifest
from tugboat.engine.types import (
    DiagnosisModel,
    FilesystemMetadata,
    HelmMetadata,
    ManifestMetadata,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterator
    from os import PathLike
    from typing import Any

logger = logging.getLogger(__name__)
yaml_parser = ruamel.yaml.YAML(typ="rt")
yaml_parser.preserve_quotes = True


def analyze_yaml_stream(
    stream: str, filepath: str | PathLike[str] | None = None
) -> list[DiagnosisModel]:
    """
    Analyze YAML manifest(s) and report issues.

    Parameters
    ----------
    stream : str
        The YAML manifest(s) as a string.
    filepath : PathLike | None
        Path to the manifest file. This is used for error reporting.

    Returns
    -------
    list[DiagnosisModel]
        The diagnoses reported by the analyzers.
    """
    digest = hashlib.md5(stream.encode()).hexdigest()
    logger.debug(f"Starting analysis YAML manifest with digest {digest}")

    metadata = None
    if filepath:
        metadata = FilesystemMetadata.model_validate(
            {
                "filepath": str(filepath),
            }
        )

    # parse the YAML stream into documents
    try:
        with io.StringIO(stream) as buf:
            if filepath:
                # set the stream name to the filepath for better error reporting
                # this is a workaround for ruamel.yaml, which does not support setting
                # the name of the stream directly.
                buf.name = str(filepath)
            documents = list(yaml_parser.load_all(buf))  # force evaluation

    except ruamel.yaml.error.MarkedYAMLError as e:
        diagnosis = translate_marked_yaml_error(e)
        diagnosis.extras.file = metadata
        return [diagnosis]

    # analyze each document
    diagnoses: list[DiagnosisModel] = []
    for i, doc in enumerate(documents, 1):
        logger.debug(f"Processing document {i}/{len(documents)} in {digest}")
        if not doc:
            continue

        if isinstance(doc, CommentedMap):
            for diagnosis in analyze_yaml_document(doc):
                diagnosis.extras.file = metadata
                diagnoses.append(diagnosis)

        elif isinstance(doc, CommentedBase):
            type_ = type(doc).__name__
            diagnoses.append(
                DiagnosisModel.model_validate(
                    {
                        "line": doc.lc.line + 1,
                        "column": doc.lc.col + 1,
                        "type": "error",
                        "code": "F003",
                        "loc": (),
                        "summary": "Malformed document structure",
                        "msg": (
                            f"Expected a YAML mapping (key-value pairs) but found a {type_} in document #{i}.\n"
                            "Kubernetes manifests must be structured as objects with properties like 'apiVersion', 'kind', etc."
                        ),
                        "extras": {
                            "file": metadata,
                        },
                    }
                )
            )

        else:
            type_ = type(doc).__name__
            diagnoses.append(
                DiagnosisModel.model_validate(
                    {
                        "line": 1,
                        "column": 1,
                        "type": "error",
                        "code": "F002",
                        "loc": (),
                        "summary": "Malformed input",
                        "msg": f"Expect a YAML mapping (key-value pairs) but found {type(doc).__name__} in document #{i}.",
                        "extras": {
                            "file": metadata,
                        },
                    }
                )
            )

    return diagnoses


def analyze_yaml_document(doc: CommentedMap) -> Iterator[DiagnosisModel]:
    """
    Analyze a single YAML document and report issues.

    This function wraps the :py:func:`analyze_manifest` function and transforms
    the :py:class:`tugboat.types.Diagnosis` objects into :py:class:`DiagnosisModel` objects.
    """
    helm_metadata = extract_helm_metadata(doc)

    for diagnosis in analyze_manifest(doc):
        diagnosis = cast("dict[str, Any]", diagnosis)
        diagnosis.setdefault("extras", {})

        if manifest_metadata := diagnosis.get("ctx", {}).get("manifest"):
            diagnosis["extras"]["manifest"] = manifest_metadata

        diagnosis["extras"]["helm"] = helm_metadata

        model = DiagnosisModel.model_validate(diagnosis)

        # TODO exclude diagnoses that are suppressed by settings

        # exclude diagnoses that are suppressed by comments
        if model.code in get_suppression_codes(doc, model.loc):
            manifest_name = "<unnamed>"
            if model.extras.manifest:
                manifest_name = model.extras.manifest.name or manifest_name
            logger.debug(
                "Diagnosis %s (%s) at %s:%s is suppressed by comment",
                model.code,
                model.summary or "<no summary>",
                manifest_name,
                model.loc_path,
            )
            continue

        # update line and column to be 1-based
        line, column = get_line_column(doc, diagnosis["loc"], diagnosis.get("input"))
        model.line = line + 1
        model.column = column + 1

        yield model


def extract_helm_metadata(doc: CommentedMap) -> dict[str, str] | None:
    """
    Extract Helm chart and template path from a YAML document's leading comment.
    """
    # find the leading comments in the document
    if not doc.ca or not doc.ca.comment:
        return None

    _, pre_comments = doc.ca.comment
    if not pre_comments:
        return None

    # it should be the first comment and start with '# Source: '
    if isinstance(fc := pre_comments[0], CommentToken) and fc.column == 0:
        if m := re.match(
            r"# Source: (?P<chart>.+?)/(?P<template>templates/.+)",
            fc.value,
        ):
            return {
                "chart": os.path.basename(m.group("chart")),
                "template": m.group("template"),
            }

    return None



from __future__ import annotations

import os

from pydantic import BaseModel, ConfigDict, Field

from tugboat.schemas.basic import Dict

if os.getenv("DOCUTILSCONFIG"):
    __all__ = ["Metadata"]


class Manifest[T_Spec: BaseModel](BaseModel):
    """
    The base schema for the Kubernetes manifest.

    This schema is generic and must be inherited with the ``T_Spec`` type
    parameter, which is the schema for the :data:`spec` field of the manifest:

    .. code-block:: python

        class MyManifestSpec(BaseModel):
            ...

        class MyManifest(Manifest[MyManifestSpec]):
            ...
    """

    model_config = ConfigDict(extra="forbid", frozen=True)

    apiVersion: str
    kind: str
    metadata: Metadata
    spec: T_Spec

    @property
    def name(self) -> str | None:
        """Short-cut to the name of the manifest."""
        return self.metadata.name or self.metadata.generateName


class Metadata(BaseModel):
    """
    Kubernetes manifest metadata.
    """

    model_config = ConfigDict(frozen=True)

    name: str | None = Field(None, min_length=1, max_length=253)
    generateName: str | None = Field(None, min_length=1, max_length=248)
    labels: Dict[str, str] | None = None
    annotations: Dict[str, str] | None = None



"""
:mod:`tugboat`, the main package of the Tugboat framework.

The Tugboat framework is a modular static analysis tool for software manifests.
It is designed to examine software manifests and report issues detected by the
analyzers.

This package exposes only the core components of the framework:

- :class:`Diagnosis`: The primary structure for a diagnosis.
- :class:`Field`: A special type to indicate a field in the manifest.
- :data:`hookimpl`: The hook implementation marker for the Tugboat framework.

Other components are available in the subpackages of the framework but are not
automatically imported here; they are intended to be imported explicitly by
the developers as needed.
"""

__all__ = [
    "Diagnosis",
    "Field",
    "__version__",
    "hookimpl",
]

from tugboat.core import hookimpl
from tugboat.types import Diagnosis, Field
from tugboat.version import __version__



"""
This module defines the core hook specifications for Tugboat.

These hooks are used to extend the core functionality of Tugboat. Since the
standard Argo Workflows manifest formats are built-in, users may not need to
implement these hooks. However, these hooks are available for users who want to
extend Tugboat to support custom manifest formats or provide custom diagnoses.
"""

from __future__ import annotations

import typing

import pluggy

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.schemas import Manifest
    from tugboat.types import Diagnosis

hookspec = pluggy.HookspecMarker("tugboat")


@hookspec(firstresult=True)
def parse_manifest(manifest: dict) -> Manifest | None:  # type: ignore[reportReturnType]
    """
    Convert a raw manifest into a :py:class:`tugboat.schemas.Manifest` instance
    if the format is recognized by the handler. Returns :py:obj:`None` if the
    format is not recognized.

    During this hook, the framework captures the :py:exc:`pydantic_core.ValidationError`
    raised when parsing the manifest and converts it into a diagnosis. This allows
    the handler to focus on the manifest structure and content, and leave the
    validation to Pydantic and the framework.

    Parameters
    ----------
    manifest : dict
        The manifest to parse.

    Returns
    -------
    Manifest
        The parsed Manifest object if the format is recognized, otherwise :py:obj:`None`.
    """


@hookspec()
def analyze(manifest: Manifest) -> Iterable[Diagnosis]:  # type: ignore[reportReturnType]
    """
    Analyze the manifest and provide diagnoses.

    This is the primary hook invoked during the linting process. The handler
    should examine the manifest and return a list of diagnoses.

    For Argo Workflows manifests, you can implement the hooks in the
    :py:mod:`tugboat.hookspecs.workflow` module. These hooks are wrapped by
    tugboat's built-in handlers, reducing the need to manage the workflow structure.

    Parameters
    ----------
    manifest : Manifest
        The manifest to analyze.

    Returns
    -------
    Iterable[Diagnosis]
        The diagnoses reported by the handler.
    """



"""
    pygments.token
    ~~~~~~~~~~~~~~

    Basic token types and the standard tokens.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""


class _TokenType(tuple):
    parent = None

    def split(self):
        buf = []
        node = self
        while node is not None:
            buf.append(node)
            node = node.parent
        buf.reverse()
        return buf

    def __init__(self, *args):
        # no need to call super.__init__
        self.subtypes = set()

    def __contains__(self, val):
        return self is val or (type(val) is self.__class__ and val[: len(self)] == self)

    def __getattr__(self, val):
        if not val or not val[0].isupper():
            return tuple.__getattribute__(self, val)
        new = _TokenType(self + (val,))
        setattr(self, val, new)
        self.subtypes.add(new)
        new.parent = self
        return new

    def __repr__(self):
        return "Token" + (self and "." or "") + ".".join(self)

    def __copy__(self):
        # These instances are supposed to be singletons
        return self

    def __deepcopy__(self, memo):
        # These instances are supposed to be singletons
        return self


Token = _TokenType()

# Special token types
Text = Token.Text
Whitespace = Text.Whitespace
Escape = Token.Escape
Error = Token.Error
# Text that doesn't belong to this lexer (e.g. HTML in PHP)
Other = Token.Other

# Common token types for source code
Keyword = Token.Keyword
Name = Token.Name
Literal = Token.Literal
String = Literal.String
Number = Literal.Number
Punctuation = Token.Punctuation
Operator = Token.Operator
Comment = Token.Comment

# Generic types for non-source code
Generic = Token.Generic

# String and some others are not direct children of Token.
# alias them:
Token.Token = Token
Token.String = String
Token.Number = Number


def is_token_subtype(ttype, other):
    """
    Return True if ``ttype`` is a subtype of ``other``.

    exists for backwards compatibility. use ``ttype in other`` now.
    """
    return ttype in other


def string_to_tokentype(s):
    """
    Convert a string into a token type::

        >>> string_to_token('String.Double')
        Token.Literal.String.Double
        >>> string_to_token('Token.Literal.Number')
        Token.Literal.Number
        >>> string_to_token('')
        Token

    Tokens that are already tokens are returned unchanged:

        >>> string_to_token(String)
        Token.Literal.String
    """
    if isinstance(s, _TokenType):
        return s
    if not s:
        return Token
    node = Token
    for item in s.split("."):
        node = getattr(node, item)
    return node


# Map standard token types to short names, used in CSS class naming.
# If you add a new item, please be sure to run this file to perform
# a consistency check for duplicate values.
STANDARD_TYPES = {
    Token: "",
    Text: "",
    Whitespace: "w",
    Escape: "esc",
    Error: "err",
    Other: "x",
    Keyword: "k",
    Keyword.Constant: "kc",
    Keyword.Declaration: "kd",
    Keyword.Namespace: "kn",
    Keyword.Pseudo: "kp",
    Keyword.Reserved: "kr",
    Keyword.Type: "kt",
    Name: "n",
    Name.Attribute: "na",
    Name.Builtin: "nb",
    Name.Builtin.Pseudo: "bp",
    Name.Class: "nc",
    Name.Constant: "no",
    Name.Decorator: "nd",
    Name.Entity: "ni",
    Name.Exception: "ne",
    Name.Function: "nf",
    Name.Function.Magic: "fm",
    Name.Property: "py",
    Name.Label: "nl",
    Name.Namespace: "nn",
    Name.Other: "nx",
    Name.Tag: "nt",
    Name.Variable: "nv",
    Name.Variable.Class: "vc",
    Name.Variable.Global: "vg",
    Name.Variable.Instance: "vi",
    Name.Variable.Magic: "vm",
    Literal: "l",
    Literal.Date: "ld",
    String: "s",
    String.Affix: "sa",
    String.Backtick: "sb",
    String.Char: "sc",
    String.Delimiter: "dl",
    String.Doc: "sd",
    String.Double: "s2",
    String.Escape: "se",
    String.Heredoc: "sh",
    String.Interpol: "si",
    String.Other: "sx",
    String.Regex: "sr",
    String.Single: "s1",
    String.Symbol: "ss",
    Number: "m",
    Number.Bin: "mb",
    Number.Float: "mf",
    Number.Hex: "mh",
    Number.Integer: "mi",
    Number.Integer.Long: "il",
    Number.Oct: "mo",
    Operator: "o",
    Operator.Word: "ow",
    Punctuation: "p",
    Punctuation.Marker: "pm",
    Comment: "c",
    Comment.Hashbang: "ch",
    Comment.Multiline: "cm",
    Comment.Preproc: "cp",
    Comment.PreprocFile: "cpf",
    Comment.Single: "c1",
    Comment.Special: "cs",
    Generic: "g",
    Generic.Deleted: "gd",
    Generic.Emph: "ge",
    Generic.Error: "gr",
    Generic.Heading: "gh",
    Generic.Inserted: "gi",
    Generic.Output: "go",
    Generic.Prompt: "gp",
    Generic.Strong: "gs",
    Generic.Subheading: "gu",
    Generic.EmphStrong: "ges",
    Generic.Traceback: "gt",
}



from __future__ import annotations

import decimal
import operator
import os
import re
from typing import TYPE_CHECKING, Any, Literal

import pydantic_core.core_schema
from pydantic import BaseModel, ConfigDict, Field

from tugboat.schemas.basic import Array
from tugboat.schemas.template.env import EnvFromSource, EnvVar
from tugboat.schemas.template.probe import Probe
from tugboat.schemas.template.volume import VolumeMount

if TYPE_CHECKING:
    from collections.abc import Callable
    from decimal import Decimal

    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema

if os.getenv("DOCUTILSCONFIG"):
    __all__ = [
        "ContainerSetRetryStrategy",
        "ContainerSetTemplate",
        "Quantity",
        "ResourceClaim",
        "ResourceQuantities",
        "ResourceRequirements",
    ]


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class _ContainerEntry(_BaseModel):

    command: Array[str] | None = None
    env: Array[EnvVar] | None = None
    envFrom: Array[EnvFromSource] | None = None
    image: str
    imagePullPolicy: Literal["Always", "Never", "IfNotPresent"] | None = None
    livenessProbe: Probe | None = None
    name: str | None = None
    readinessProbe: Probe | None = None
    resources: ResourceRequirements | None = None
    restartPolicy: Literal["Always"] | None = None
    startupProbe: Probe | None = None
    stdin: bool | None = None
    stdinOnce: bool | None = None
    terminationMessagePath: str | None = None
    terminationMessagePolicy: Literal["File", "FallbackToLogsOnError"] | None = None
    tty: bool | None = None
    volumeMounts: Array[VolumeMount] | None = None
    workingDir: str | None = None

    lifecycle: Any | None = None
    ports: Array[Any] | None = None
    resizePolicy: Array[Any] | None = None
    securityContext: Any | None = None
    volumeDevices: Array[Any] | None = None


class ContainerTemplate(_ContainerEntry):
    """
    A single application `container`_ that you want to run within a pod.

    .. _container: https://argo-workflows.readthedocs.io/en/latest/fields/#container
    """

    args: Array[str] | None = None

    def __hash__(self):
        return hash((self.image, self.command, self.args))


class ScriptTemplate(_ContainerEntry):
    """
    `ScriptTemplate`_ is a template subtype to enable scripting through code steps.

    .. _ScriptTemplate: https://argo-workflows.readthedocs.io/en/latest/fields/#scripttemplate
    """

    source: str

    def __hash__(self):
        return hash((self.image, self.source))


class ContainerNode(_ContainerEntry):
    """
    Represents an individual `ContainerNode`_ within a `ContainerSetTemplate`_.

    .. _ContainerNode:
       https://argo-workflows.readthedocs.io/en/latest/fields/#containernode
    .. _ContainerSetTemplate:
       https://argo-workflows.readthedocs.io/en/latest/fields/#containersettemplate
    """

    args: Array[str] | None = None
    dependencies: Array[str] | None = None

    def __hash__(self):
        return hash((self.image, self.command, self.args))


class ContainerSetTemplate(_BaseModel):
    """
    `ContainerSetTemplate`_ to specify multiple containers to run within a single pod.

    .. _ContainerSetTemplate:
       https://argo-workflows.readthedocs.io/en/latest/fields/#containersettemplate
    """

    containers: Array[ContainerNode]
    retryStrategy: ContainerSetRetryStrategy | None = None
    volumeMounts: Array[VolumeMount] | None = None


class ContainerSetRetryStrategy(_BaseModel):
    duration: str | None = Field(None, pattern=r"\d+(ns|us|µs|ms|s|m|h)")
    retries: int | str


class ResourceRequirements(_BaseModel):
    """
    `ResourceRequirements`_ describes the compute resource requirements.

    .. _ResourceRequirements:
       https://argo-workflows.readthedocs.io/en/latest/fields/#resourcerequirements
    """

    claims: Array[ResourceClaim] | None = None
    limits: ResourceQuantities | None = None
    requests: ResourceQuantities | None = None


class ResourceClaim(_BaseModel):
    """
    `ResourceClaim`_ references one entry in PodSpec.ResourceClaims.

    .. _ResourceClaim: https://argo-workflows.readthedocs.io/en/latest/fields/#resourceclaim
    """

    name: str
    request: str | None = None


class ResourceQuantities(_BaseModel):
    """
    ResourceQuantity is a class to represent resource quantities in Kubernetes.
    """

    cpu: Quantity | None = None
    memory: Quantity | None = None


class Quantity:
    """
    `Quantity`_ is a fixed-point representation of a number.

    .. _Quantity: https://kubernetes.io/docs/reference/kubernetes-api/common-definitions/quantity/#Quantity
    """

    expr: str
    value: decimal.Decimal

    def __init__(self, expr: str) -> None:
        expr = expr.strip()

        if m := re.search(r"(Ki|Mi|Gi|Ti|Pi|Ei|m|k|M|G|T|P|E)$", expr):
            num = expr[: m.start()]
            suffix = m.group(0)

            multiplier = {
                "Ki": 1024,
                "Mi": 1024**2,
                "Gi": 1024**3,
                "Ti": 1024**4,
                "Pi": 1024**5,
                "Ei": 1024**6,
                "m": decimal.Decimal("0.001"),
                "k": 1000,
                "M": 1_000_000,
                "G": 1_000_000_000,
                "T": 1_000_000_000_000,
                "P": 1_000_000_000_000_000,
                "E": 1_000_000_000_000_000_000,
            }[suffix]

        else:
            num = expr
            multiplier = 1

        try:
            dec = decimal.Decimal(num)
        except decimal.InvalidOperation:
            raise ValueError("Invalid decimal quantity") from None

        dec *= multiplier
        assert isinstance(dec, decimal.Decimal)  # satisfy type checker

        # it may not have more than 3 decimal places
        dec = dec.quantize(decimal.Decimal("0.001"), rounding=decimal.ROUND_UP)

        # this class is only used for cpu and memory quantities, which must be non-negative
        if dec < 0:
            raise ValueError("Quantity must be non-negative")

        self.expr = expr
        self.value = dec

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:

        def _validator(v: Any):
            if isinstance(v, int | float):
                v = str(v)
            if isinstance(v, str):
                v = cls(v)
            return v

        python_schema = pydantic_core.core_schema.no_info_before_validator_function(
            _validator, pydantic_core.core_schema.is_instance_schema(cls)
        )
        json_schema = pydantic_core.core_schema.no_info_after_validator_function(
            str, pydantic_core.core_schema.str_schema()
        )
        return pydantic_core.core_schema.json_or_python_schema(
            python_schema=python_schema,
            json_schema=json_schema,
            serialization=pydantic_core.core_schema.to_string_ser_schema(),
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.expr})"

    def __str__(self) -> str:
        return self.expr

    def __hash__(self) -> int:
        return hash(self.value)

    def _compare(self, other: Any, method: Callable[[Decimal, Decimal], bool]) -> bool:
        if isinstance(other, type(self)):
            return method(self.value, other.value)
        raise TypeError

    def __lt__(self, other) -> bool:
        return self._compare(other, operator.lt)

    def __le__(self, other) -> bool:
        return self._compare(other, operator.le)

    def __eq__(self, other) -> bool:
        return self._compare(other, operator.eq)

    def __ge__(self, other) -> bool:
        return self._compare(other, operator.ge)

    def __gt__(self, other) -> bool:
        return self._compare(other, operator.gt)

    def __ne__(self, other) -> bool:
        return self._compare(other, operator.ne)



from __future__ import annotations

import collections
import json
import typing
from collections.abc import Mapping, Sequence

from rapidfuzz.process import extractOne

from tugboat.types import Field
from tugboat.utils.humanize import get_context_name, join_with_or

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any

    from pydantic_core import ErrorDetails

    from tugboat.types import Bundle, Diagnosis


def bulk_translate_pydantic_errors(
    errors: Iterable[ErrorDetails],
) -> list[Diagnosis]:
    """
    Translate multiple Pydantic errors to diagnosis objects.

    This function helps to translate multiple Pydantic errors to diagnosis objects,
    and merges some of the similar errors into a more concise message.

    See :func:`translate_pydantic_error` for the details of the translation.

    Parameters
    ----------
    errors : ~collections.abc.Iterable[~pydantic_core.ErrorDetails]
        An iterable of Pydantic error objects. These objects could be obtained from
        :py:meth:`ValidationError.errors <pydantic_core.ValidationError.errors>` method.

    Yields
    ------
    list[Diagnosis]
        A list of diagnosis objects that contain the error messages
    """
    diagnoes = []

    # for union types, pydantic raises multiple errors for the same field
    # so handle them separately

    def _is_union_type_error(err: ErrorDetails, tp: str) -> bool:
        """
        filter error related to union types

        we observed that pydantic appends the type name to the 'loc' for union type:

        ```json
        {
            "type": "int_type",
            "loc": ["foo", "bar", "int"],
            "msg": "Input should be a valid integer"
        }
        ```
        """
        return err["loc"][-1] == tp and (
            err["type"] == f"{tp}_type" or err["type"] == f"{tp}_parsing"
        )

    union_errors = collections.defaultdict(list)
    for err in errors:
        if (
            False
            or _is_union_type_error(err, "bool")
            or _is_union_type_error(err, "bytes")
            or _is_union_type_error(err, "date")
            or _is_union_type_error(err, "datetime")
            or _is_union_type_error(err, "decimal")
            or _is_union_type_error(err, "dict")
            or _is_union_type_error(err, "float")
            or _is_union_type_error(err, "frozen_set")
            or _is_union_type_error(err, "int")
            or _is_union_type_error(err, "iterable")
            or _is_union_type_error(err, "json")
            or _is_union_type_error(err, "list")
            or _is_union_type_error(err, "mapping")
            or _is_union_type_error(err, "set")
            or _is_union_type_error(err, "time_delta")
            or _is_union_type_error(err, "time")
            or _is_union_type_error(err, "tuple")
            or _is_union_type_error(err, "url")
            or _is_union_type_error(err, "uuid")
        ):
            union_errors[err["loc"][:-1]].append(err)

        elif err["type"] == "string_type" and err["loc"][-1] == "str":  # :,)
            union_errors[err["loc"][:-1]].append(err)

        else:
            # other errors - yield as is
            diagnoes.append(translate_pydantic_error(err))

    # merge union type errors
    for loc, errors in union_errors.items():
        # collect all expected types
        expected_types = {err["loc"][-1] for err in errors}
        expected_type_expr = join_with_or(expected_types, quote=False)

        # build a more concise message
        _, field = _get_field_name(loc)
        input_type = get_type_name(errors[0]["input"])

        diagnoes.append(
            {
                "type": "failure",
                "code": "M103",
                "loc": loc,
                "summary": "Input type mismatch",
                "msg": (
                    f"Expected {expected_type_expr} for field {field}, but received a {input_type}."
                ),
                "input": errors[0]["input"],
            }
        )

    return diagnoes


def translate_pydantic_error(error: ErrorDetails) -> Diagnosis:  # noqa: C901
    """
    Translate a Pydantic error to a diagnosis object.

    This function returns a diagnosis object based on the error type found in the input:

    .. list-table::

        * - Pydantic Error Type
          - Tugboat Code
        * - `bool_parsing <https://docs.pydantic.dev/latest/errors/validation_errors/#bool_parsing>`_
          - :rule:`m103`
        * - `bool_type <https://docs.pydantic.dev/latest/errors/validation_errors/#bool_type>`_
          - :rule:`m103`
        * - `dict_type <https://docs.pydantic.dev/latest/errors/validation_errors/#dict_type>`_
          - :rule:`m103`
        * - `decimal_parsing <https://docs.pydantic.dev/latest/errors/validation_errors/#decimal_parsing>`_
          - :rule:`m103`
        * - `decimal_type <https://docs.pydantic.dev/latest/errors/validation_errors/#decimal_type>`_
          - :rule:`m103`
        * - `enum <https://docs.pydantic.dev/latest/errors/validation_errors/#enum>`_
          - :rule:`m104`
        * - `extra_forbidden <https://docs.pydantic.dev/latest/errors/validation_errors/#extra_forbidden>`_
          - :rule:`m102`
        * - `float_parsing <https://docs.pydantic.dev/latest/errors/validation_errors/#float_parsing>`_
          - :rule:`m103`
        * - `float_type <https://docs.pydantic.dev/latest/errors/validation_errors/#float_type>`_
          - :rule:`m103`
        * - `frozen_set_type <https://docs.pydantic.dev/latest/errors/validation_errors/#frozen_set_type>`_
          - :rule:`m103`
        * - `int_parsing <https://docs.pydantic.dev/latest/errors/validation_errors/#int_parsing>`_
          - :rule:`m103`
        * - `int_type <https://docs.pydantic.dev/latest/errors/validation_errors/#int_type>`_
          - :rule:`m103`
        * - `iterable_type <https://docs.pydantic.dev/latest/errors/validation_errors/#iterable_type>`_
          - :rule:`m103`
        * - `list_type <https://docs.pydantic.dev/latest/errors/validation_errors/#list_type>`_
          - :rule:`m103`
        * - `literal_error <https://docs.pydantic.dev/latest/errors/validation_errors/#literal_error>`_
          - :rule:`m104`
        * - `mapping_type <https://docs.pydantic.dev/latest/errors/validation_errors/#mapping_type>`_
          - :rule:`m103`
        * - `missing <https://docs.pydantic.dev/latest/errors/validation_errors/#missing>`_
          - :rule:`m101`
        * - `set_type <https://docs.pydantic.dev/latest/errors/validation_errors/#set_type>`_
          - :rule:`m103`
        * - `string_type <https://docs.pydantic.dev/latest/errors/validation_errors/#string_type>`_
          - :rule:`m103`
        * - `tuple_type <https://docs.pydantic.dev/latest/errors/validation_errors/#tuple_type>`_
          - :rule:`m103`
        * - Any other error
          - :rule:`m003`

    Below are some custom error types defined in Tugboat:

    .. list-table::

        * - Custom Error Type
          - Tugboat Code
        * - ``artifact_prohibited_value_field``
          - :rule:`m102`
        * - ``parameter_value_type_error``
          - :rule:`m103`

    Parameters
    ----------
    error : ~pydantic_core.ErrorDetails
        A Pydantic error object. This object could be obtained from
        :py:meth:`ValidationError.errors <pydantic_core.ValidationError.errors>` method.

    Returns
    -------
    ~tugboat.types.Diagnosis
       A diagnosis object that contains the error message and other relevant information.
    """
    match error["type"]:
        case "bool_parsing" | "bool_type":
            _, field = _get_field_name(error["loc"])
            input_type = get_type_name(error["input"])
            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid boolean",
                "msg": (
                    f"Expected a boolean for field {field}, but received a {input_type}.\n"
                    "Try using 'true' or 'false' without quotes."
                ),
                "input": error["input"],
            }

        case "dict_type" | "mapping_type":
            _, field = _get_field_name(error["loc"])
            input_type = get_type_name(error["input"])

            if not error["input"]:
                return {
                    "type": "failure",
                    "code": "M103",
                    "loc": error["loc"],
                    "summary": "Input should be a valid mapping",
                    "msg": (
                        f"Expected a mapping for field {field}, but received a {input_type}.\n"
                        "If an empty mapping is intended, use '{}'."
                    ),
                    "input": error["input"],
                    "fix": "{}",
                }

            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid mapping",
                "msg": f"Expected a mapping for field {field}, but received a {input_type}.",
                "input": error["input"],
            }

        case "decimal_parsing" | "decimal_type" | "float_parsing" | "float_type":
            _, field = _get_field_name(error["loc"])
            input_type = get_type_name(error["input"])
            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid number",
                "msg": f"Expected a number for field {field}, but received a {input_type}.",
                "input": error["input"],
            }

        case "enum" | "literal_error":
            expected_literal = error.get("ctx", {}).get("expected", "")
            expected = _extract_expects(expected_literal)

            _, field = _get_field_name(error["loc"])

            fix = None
            if result := extractOne(input_ := error["input"], expected):
                fix, _, _ = result

            return {
                "type": "failure",
                "code": "M104",
                "loc": error["loc"],
                "summary": error["msg"],
                "msg": (
                    f"Input '{input_}' is not a valid value for field {field}.\n"
                    f"Expected {expected_literal}."
                ),
                "input": error["input"],
                "fix": fix,
            }

        case "extra_forbidden":
            raw_field_name, formatted_field = _get_field_name(error["loc"])
            *parents, _ = error["loc"]
            return {
                "type": "failure",
                "code": "M102",
                "loc": error["loc"],
                "summary": "Found redundant field",
                "msg": f"Field {formatted_field} is not valid within {get_context_name(parents)}.",
                "input": Field(raw_field_name),
            }

        case (
            "frozen_set_type"
            | "iterable_type"
            | "list_type"
            | "set_type"
            | "tuple_type"
        ):
            _, field = _get_field_name(error["loc"])
            input_type = get_type_name(error["input"])

            if not error["input"]:
                return {
                    "type": "failure",
                    "code": "M103",
                    "loc": error["loc"],
                    "summary": "Input should be a valid array",
                    "msg": (
                        f"Expected an array for field {field}, but received a {input_type}.\n"
                        "If an empty array is intended, use '[]'."
                    ),
                    "input": error["input"],
                    "fix": "[]",
                }

            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid array",
                "msg": f"Expected an array for field {field}, but received a {input_type}.",
                "input": error["input"],
            }

        case "int_parsing" | "int_type":
            _, field = _get_field_name(error["loc"])
            input_type = get_type_name(error["input"])
            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid integer",
                "msg": f"Expected a integer for field {field}, but received a {input_type}.",
                "input": error["input"],
            }

        case "missing":
            _, field = _get_field_name(error["loc"])
            return {
                "type": "failure",
                "code": "M101",
                "loc": error["loc"],
                "summary": "Missing required field",
                "msg": f"Field {field} is required but missing",
            }

        case "string_type":
            _, field = _get_field_name(error["loc"])
            return {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a valid string",
                "msg": "\n".join(_compose_string_error_message(field, error["input"])),
                "input": error["input"],
            }

        case "artifact_prohibited_value_field":
            diagnosis: Diagnosis = {
                "type": "failure",
                "code": "M102",
                "loc": error["loc"],
                "summary": "Invalid field for artifact",
                "msg": error["msg"],
                "input": Field("value"),
            }

            try:
                diagnosis["fix"] = json.dumps({"raw": {"data": error["input"]}})
            except Exception:
                ...

            return diagnosis

        case "parameter_value_type_error":
            input_type = get_type_name(error["input"])

            diagnosis: Diagnosis = {
                "type": "failure",
                "code": "M103",
                "loc": error["loc"],
                "summary": "Input should be a string",
                "msg": (
                    f"Expected string for parameter value, but received a {input_type}."
                ),
                "input": error["input"],
            }

            if isinstance(error["input"], dict | list):
                try:
                    diagnosis["fix"] = json.dumps(error["input"], indent=2)
                except Exception:
                    ...

                diagnosis["msg"] += (
                    "\n"
                    "If a complex structure is intended, serialize it as a JSON string."
                )

            return diagnosis

    return {
        "type": "failure",
        "code": "M003",
        "loc": error["loc"],
        "msg": error["msg"],
        "input": error["input"],
        "ctx": {
            "pydantic_error": typing.cast("Bundle", error),
        },
    }


def get_type_name(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, bool):
        return "boolean"
    if isinstance(value, int):
        return "integer"
    if isinstance(value, str):
        return "string"
    if isinstance(value, float):
        return "number"
    if isinstance(value, Mapping):
        return "mapping"
    if isinstance(value, Sequence):
        return "array"
    return type(value).__name__


def _get_field_name(loc: tuple[int | str, ...]) -> tuple[str | None, str]:
    """
    Get the last string in the location tuple as the field name.

    Returns
    -------
    raw : str
        The raw field name.
    quoted : str
        The quoted field name for display.
    """
    for item in reversed(loc):
        if isinstance(item, str):
            return item, f"'{item}'"
    return None, "<unnamed>"


def _extract_expects(literal: str) -> Iterator[str]:
    """
    Extract the expected values from pydantic's error message.

    The expected value string is like:

    .. code-block:: none

       "hello'", 'world' or 'hola'
    """
    idx = 0
    while idx < len(literal):
        if literal[idx] == "'":
            idx_end = literal.find("'", idx + 1)
            yield literal[idx + 1 : idx_end]
            idx = idx_end + 1

        elif literal[idx] == '"':
            idx_end = literal.find('"', idx + 1)
            yield literal[idx + 1 : idx_end]
            idx = idx_end + 1

        else:
            idx += 1


def _compose_string_error_message(field: str, value: Any) -> Iterator[str]:
    """
    Construct an error message for string type validation.
    Includes user suggestions based on the value and common YAML parsing pitfalls.

    Ref: https://ruudvanasseldonk.com/2023/01/11/the-yaml-document-from-hell
    """
    input_type = get_type_name(value)
    yield f"Expected a string for field {field}, but received a {input_type}."

    # the Norway problem
    if isinstance(value, bool):
        if value is True:
            yield "Note that these inputs will be interpreted as boolean true: 'True', 'Yes', 'On', 'Y'."
        else:
            yield "Note that these inputs will be interpreted as boolean false: 'False', 'No', 'Off', 'N'."

    # sexagesimal
    if isinstance(value, int) and 60 < value <= 3600:
        sexagesimal = _to_sexagesimal(value)
        yield (
            f"Numbers separated by colons (e.g. {sexagesimal}) will be interpreted as sexagesimal."
        )

    # general suggestion
    yield "Try using quotes for strings to fix this issue."


def _to_sexagesimal(value: int) -> str:
    """Convert an integer to a sexagesimal string."""
    if value < 0:
        sign = "-"
        value = -value
    else:
        sign = ""

    digits = []
    while value:
        digits.append(value % 60)
        value //= 60

    return sign + ":".join(str(d) for d in reversed(digits))
